import base64
import html
import io
import json
import logging
import mimetypes
from typing import Any, Callable, Optional

import av
import numpy as np
from PIL import Image

import pixeltable.type_system as ts
from pixeltable.utils.http_server import get_file_uri

_logger = logging.getLogger('pixeltable')


class Formatter:
    """
    A factory for constructing HTML formatters for Pixeltable data. The formatters are used to customize
    the rendering of `DataFrameResultSet`s in notebooks.

    Args:
        num_rows: Number of rows in the DataFrame being rendered.
        num_cols: Number of columns in the DataFrame being rendered.
        http_address: Root address of the Pixeltable HTTP server (used to construct URLs for media references).
    """

    __FLOAT_PRECISION = 3
    __LIST_THRESHOLD = 16
    __LIST_EDGEITEMS = 6
    __STRING_SEP = ' ...... '
    __STRING_MAX_LEN = 1000
    __NESTED_STRING_MAX_LEN = 300

    def __init__(self, num_rows: int, num_cols: int, http_address: str):
        self.__num_rows = num_rows
        self.__num_cols = num_cols
        self.__http_address = http_address

    def get_pandas_formatter(self, col_type: ts.ColumnType) -> Optional[Callable]:
        if col_type.is_string_type():
            return self.format_string
        if col_type.is_float_type():
            return self.format_float
        if col_type.is_json_type():
            return self.format_json
        if col_type.is_array_type():
            return self.format_array
        if col_type.is_image_type():
            return self.format_img
        if col_type.is_video_type():
            return self.format_video
        if col_type.is_audio_type():
            return self.format_audio
        if col_type.is_document_type():
            return self.format_document
        return None

    @classmethod
    def format_string(cls, val: str) -> str:
        """
        Escapes special characters in `val`, and abbreviates `val` if its length exceeds `_STRING_MAX_LEN`.
        """
        return cls.__escape(cls.abbreviate(val))

    @classmethod
    def abbreviate(cls, val: str, max_len: int = __STRING_MAX_LEN) -> str:
        if len(val) > max_len:
            edgeitems = (max_len - len(cls.__STRING_SEP)) // 2
            return f'{val[:edgeitems]}{cls.__STRING_SEP}{val[-edgeitems:]}'
        return val

    @classmethod
    def __escape(cls, val: str) -> str:
        # HTML-escape the specified string, then escape $ signs to suppress MathJax formatting
        # TODO(aaron-siegel): The '$' escaping isn't perfect; it will fail on '$' that are already escaped
        return html.escape(val).replace('$', r'\$')

    @classmethod
    def format_float(cls, val: float) -> str:
        # stay consistent with numpy formatting (0-D array has no brackets)
        return np.array2string(np.array(val), precision=cls.__FLOAT_PRECISION)

    @classmethod
    def format_array(cls, arr: np.ndarray) -> str:
        return np.array2string(
            arr,
            precision=cls.__FLOAT_PRECISION,
            threshold=cls.__LIST_THRESHOLD,
            edgeitems=cls.__LIST_EDGEITEMS,
            max_line_width=1000000,
        )

    @classmethod
    def format_json(cls, val: Any, escape_strings: bool = True) -> str:
        if isinstance(val, str):
            # JSON-like formatting will be applied to strings that appear nested within a list or dict
            # (quote the string; escape any quotes inside the string; shorter abbreviations).
            # However, if the string appears in top-level position (i.e., the entire JSON value is a
            # string), then we format it like an ordinary string.
            return cls.format_string(val) if escape_strings else cls.abbreviate(val)
        # In all other cases, dump the JSON struct recursively.
        return cls.__format_json_rec(val, escape_strings)

    @classmethod
    def __format_json_rec(cls, val: Any, escape_strings: bool) -> str:
        if isinstance(val, str):
            formatted = json.dumps(cls.abbreviate(val, cls.__NESTED_STRING_MAX_LEN))
            return cls.__escape(formatted) if escape_strings else formatted
        if isinstance(val, float):
            return cls.format_float(val)
        if isinstance(val, np.ndarray):
            return cls.format_array(val)
        if isinstance(val, list):
            if len(val) < cls.__LIST_THRESHOLD:
                components = [cls.__format_json_rec(x, escape_strings) for x in val]
            else:
                components = [cls.__format_json_rec(x, escape_strings) for x in val[: cls.__LIST_EDGEITEMS]]
                components.append('...')
                components.extend(cls.__format_json_rec(x, escape_strings) for x in val[-cls.__LIST_EDGEITEMS :])
            return '[' + ', '.join(components) + ']'
        if isinstance(val, dict):
            kv_pairs = (
                f'{cls.__format_json_rec(k, escape_strings)}: {cls.__format_json_rec(v, escape_strings)}'
                for k, v in val.items()
            )
            return '{' + ', '.join(kv_pairs) + '}'

        # Everything else
        try:
            return json.dumps(val)
        except TypeError:  # Not JSON serializable
            return cls.__escape(str(val))

    def format_img(self, img: Image.Image) -> str:
        """
        Create <img> tag for Image object.
        """
        assert isinstance(img, Image.Image), f'Wrong type: {type(img)}'
        # Try to make it look decent in a variety of display scenarios
        if self.__num_rows > 1:
            width = min(240, img.width)  # Multiple rows: display small images
        elif self.__num_cols > 1:
            width = min(480, img.width)  # Multiple columns: display medium images
        else:
            width = min(640, img.width)  # A single image: larger display
        with io.BytesIO() as buffer:
            img.save(buffer, 'webp')
            img_base64 = base64.b64encode(buffer.getvalue()).decode()
            return f"""
            <div class="pxt_image" style="width:{width}px;">
                <img src="data:image/webp;base64,{img_base64}" width="{width}" />
            </div>
            """

    def format_video(self, file_path: str) -> str:
        # Attempt to extract the first frame of the video to use as a thumbnail,
        # so that the notebook can be exported as HTML and viewed in contexts where
        # the video itself is not accessible.
        # TODO(aaron-siegel): If the video is backed by a concrete external URL,
        # should we link to that instead?
        thumb = self.extract_first_video_frame(file_path)
        if thumb is None:
            thumb_tag = ''
        else:
            with io.BytesIO() as buffer:
                thumb.save(buffer, 'jpeg')
                thumb_base64 = base64.b64encode(buffer.getvalue()).decode()
                thumb_tag = f'poster="data:image/jpeg;base64,{thumb_base64}"'
        if self.__num_rows > 1:
            width = 320
        elif self.__num_cols > 1:
            width = 480
        else:
            width = 800
        return f"""
        <div class="pxt_video" style="width:{width}px;">
            <video controls width="{width}" {thumb_tag}>
                {self.__create_source_tag(self.__http_address, file_path)}
            </video>
        </div>
        """

    @classmethod
    def extract_first_video_frame(cls, file_path: str) -> Optional[Image.Image]:
        with av.open(file_path) as container:
            try:
                img = next(container.decode(video=0)).to_image()
                assert isinstance(img, Image.Image)
                return img
            except Exception:
                return None

    def format_audio(self, file_path: str) -> str:
        return f"""
        <div class="pxt_audio">
            <audio controls>
                {self.__create_source_tag(self.__http_address, file_path)}
            </audio>
        </div>
        """

    def format_document(self, file_path: str, max_width: int = 320, max_height: int = 320) -> str:
        # by default, file path will be shown as a link
        inner_element = file_path
        inner_element = html.escape(inner_element)

        thumb = self.make_document_thumbnail(file_path, max_width, max_height)
        if thumb is not None:
            with io.BytesIO() as buffer:
                thumb.save(buffer, 'webp')
                thumb_base64 = base64.b64encode(buffer.getvalue()).decode()
                thumb_tag = f'data:image/webp;base64,{thumb_base64}'
            inner_element = f'<img style="object-fit: contain; border: 1px solid black;" src="{thumb_tag}" />'

        return f"""
        <div class="pxt_document" style="width:{max_width}px;">
            <a href="{get_file_uri(self.__http_address, file_path)}">
                {inner_element}
            </a>
        </div>
        """

    @classmethod
    def make_document_thumbnail(
        cls, file_path: str, max_width: int = 320, max_height: int = 320
    ) -> Optional[Image.Image]:
        """
        Returns a thumbnail image of a document.
        """
        if file_path.lower().endswith('.pdf'):
            try:
                import fitz  # type: ignore[import-untyped]

                doc = fitz.open(file_path)
                pixmap = doc.get_page_pixmap(0)
                while pixmap.width > max_width or pixmap.height > max_height:
                    # shrink(1) will halve each dimension
                    pixmap.shrink(1)
                return pixmap.pil_image()
            except Exception:
                logging.warning(f'Failed to produce PDF thumbnail {file_path}. Make sure you have PyMuPDF installed.')

        return None

    @classmethod
    def __create_source_tag(cls, http_address: str, file_path: str) -> str:
        src_url = get_file_uri(http_address, file_path)
        mime = mimetypes.guess_type(src_url)[0]
        # if mime is None, the attribute string would not be valid html.
        mime_attr = f'type="{mime}"' if mime is not None else ''
        return f'<source src="{src_url}" {mime_attr} />'



from __future__ import annotations

import importlib
import inspect
import typing
from abc import ABC, abstractmethod
from copy import copy
from typing import TYPE_CHECKING, Any, Callable, Optional, Sequence, cast

import sqlalchemy as sql
from typing_extensions import Self

from pixeltable import exceptions as excs, type_system as ts

from .globals import resolve_symbol
from .signature import Signature

if TYPE_CHECKING:
    from pixeltable import exprs

    from .expr_template_function import ExprTemplate, ExprTemplateFunction


class Function(ABC):
    """Base class for Pixeltable's function interface.

    A function in Pixeltable is an object that has a signature and implements __call__().
    This base class provides a default serialization mechanism for Function instances provided by Python modules,
    via the member self_path.
    """

    signatures: list[Signature]
    self_path: Optional[str]
    is_method: bool
    is_property: bool
    _conditional_return_type: Optional[Callable[..., ts.ColumnType]]

    # We cache the overload resolutions in self._resolutions. This ensures that each resolution is represented
    # globally by a single Python object. We do this dynamically rather than pre-constructing them in order to
    # avoid circular complexity in the `Function` initialization logic.
    __resolved_fns: list[Self]

    # Translates a call to this function with the given arguments to its SQLAlchemy equivalent.
    # Overriden for specific Function instances via the to_sql() decorator. The override must accept the same
    # parameter names as the original function. Each parameter is going to be of type sql.ColumnElement.
    _to_sql: Callable[..., Optional[sql.ColumnElement]]

    # Returns the resource pool to use for calling this function with the given arguments.
    # Overriden for specific Function instances via the resource_pool() decorator. The override must accept a subset
    # of the parameters of the original function, with the same type.
    _resource_pool: Callable[..., Optional[str]]

    def __init__(
        self,
        signatures: list[Signature],
        self_path: Optional[str] = None,
        is_method: bool = False,
        is_property: bool = False,
    ):
        # Check that stored functions cannot be declared using `is_method` or `is_property`:
        assert not ((is_method or is_property) and self_path is None)
        assert isinstance(signatures, list)
        self.signatures = signatures
        self.self_path = self_path  # fully-qualified path to self
        self.is_method = is_method
        self.is_property = is_property
        self._conditional_return_type = None
        self.__resolved_fns = []
        self._to_sql = self.__default_to_sql
        self._resource_pool = self.__default_resource_pool

    @property
    def is_valid(self) -> bool:
        return len(self.signatures) > 0

    @property
    def name(self) -> str:
        assert self.self_path is not None
        return self.self_path.split('.')[-1]

    @property
    def display_name(self) -> str:
        if self.self_path is None:
            return '<anonymous>'
        ptf_prefix = 'pixeltable.functions.'
        if self.self_path.startswith(ptf_prefix):
            return self.self_path[len(ptf_prefix) :]
        return self.self_path

    @property
    def is_polymorphic(self) -> bool:
        return len(self.signatures) > 1

    @property
    def signature(self) -> Signature:
        assert not self.is_polymorphic
        return self.signatures[0]

    @property
    def arity(self) -> int:
        assert not self.is_polymorphic
        return len(self.signature.parameters)

    @property
    @abstractmethod
    def is_async(self) -> bool: ...

    def comment(self) -> Optional[str]:
        return None

    def help_str(self) -> str:
        docstring = self.comment()
        display = self.display_name + str(self.signatures[0])
        if docstring is None:
            return display
        return f'{display}\n\n{docstring}'

    @property
    def _resolved_fns(self) -> list[Self]:
        """
        Return the list of overload resolutions for this `Function`, constructing it first if necessary.
        Each resolution is a new `Function` instance that retains just the single signature at index `signature_idx`,
        and is otherwise identical to this `Function`.
        """
        if len(self.__resolved_fns) == 0:
            # The list of overload resolutions hasn't been constructed yet; do so now.
            if len(self.signatures) == 1:
                # Only one signature: no need to construct separate resolutions
                self.__resolved_fns.append(self)
            else:
                # Multiple signatures: construct a resolution for each signature
                for idx in range(len(self.signatures)):
                    resolution = cast(Self, copy(self))
                    resolution.signatures = [self.signatures[idx]]
                    resolution.__resolved_fns = [resolution]  # Resolves to itself
                    resolution._update_as_overload_resolution(idx)
                    self.__resolved_fns.append(resolution)

        return self.__resolved_fns

    @property
    def _has_resolved_fns(self) -> bool:
        """
        Returns true if the resolved_fns for this `Function` have been constructed (i.e., if self._resolved_fns
        has been accessed).
        """
        return len(self.__resolved_fns) > 0

    def _update_as_overload_resolution(self, signature_idx: int) -> None:
        """
        Subclasses must implement this in order to do any additional work when creating a resolution, beyond
        simply updating `self.signatures`.
        """
        raise NotImplementedError()

    def __call__(self, *args: Any, **kwargs: Any) -> 'exprs.FunctionCall':
        from pixeltable import exprs

        args = [exprs.Expr.from_object(arg) for arg in args]
        kwargs = {k: exprs.Expr.from_object(v) for k, v in kwargs.items()}

        for i, expr in enumerate(args):
            if expr is None:
                raise excs.Error(f'Argument {i + 1} in call to {self.self_path!r} is not a valid Pixeltable expression')
        for param_name, expr in kwargs.items():
            if expr is None:
                raise excs.Error(
                    f'Argument {param_name!r} in call to {self.self_path!r} is not a valid Pixeltable expression'
                )

        resolved_fn, bound_args = self._bind_to_matching_signature(args, kwargs)
        return_type = resolved_fn.call_return_type(bound_args)

        return exprs.FunctionCall(resolved_fn, args, kwargs, return_type)

    def _bind_to_matching_signature(self, args: Sequence[Any], kwargs: dict[str, Any]) -> tuple[Self, dict[str, Any]]:
        result: int = -1
        bound_args: Optional[dict[str, Any]] = None
        assert len(self.signatures) > 0
        if len(self.signatures) == 1:
            # Only one signature: call _bind_to_signature() and surface any errors directly
            result = 0
            bound_args = self._bind_to_signature(0, args, kwargs)
        else:
            # Multiple signatures: try each signature in declaration order and trap any errors.
            # If none of them succeed, raise a generic error message.
            for i in range(len(self.signatures)):
                try:
                    bound_args = self._bind_to_signature(i, args, kwargs)
                except (TypeError, excs.Error):
                    continue
                result = i
                break
            if result == -1:
                raise excs.Error(f'Function {self.name!r} has no matching signature for arguments')
        assert result >= 0
        assert bound_args is not None
        return self._resolved_fns[result], bound_args

    def _bind_to_signature(self, signature_idx: int, args: Sequence[Any], kwargs: dict[str, Any]) -> dict[str, Any]:
        from pixeltable import exprs

        signature = self.signatures[signature_idx]
        bound_args = signature.py_signature.bind(*args, **kwargs).arguments
        normalized_args = {k: exprs.Expr.from_object(v) for k, v in bound_args.items()}
        self._resolved_fns[signature_idx].validate_call(normalized_args)
        return normalized_args

    def validate_call(self, bound_args: dict[str, Optional['exprs.Expr']]) -> None:
        """Override this to do custom validation of the arguments"""
        assert not self.is_polymorphic
        self.signature.validate_args(bound_args, context=f'in function {self.name!r}')

    def call_resource_pool(self, bound_args: dict[str, 'exprs.Expr']) -> str:
        """Return the resource pool to use for calling this function with the given arguments"""
        rp_kwargs = self._assemble_callable_args(self._resource_pool, bound_args)
        if rp_kwargs is None:
            # TODO: What to do in this case? An example where this can happen is if model_id is not a constant
            #   in a call to one of the OpenAI endpoints.
            raise excs.Error('Could not determine resource pool')
        return self._resource_pool(**rp_kwargs)

    def call_return_type(self, bound_args: dict[str, 'exprs.Expr']) -> ts.ColumnType:
        """Return the type of the value returned by calling this function with the given arguments"""
        if self._conditional_return_type is None:
            # No conditional return type specified; use the default return type
            return_type = self.signature.return_type
        else:
            crt_kwargs = self._assemble_callable_args(self._conditional_return_type, bound_args)
            if crt_kwargs is None:
                # A conditional return type is specified, but one of its arguments is not a constant.
                # Use the default return type
                return_type = self.signature.return_type
            else:
                # A conditional return type is specified and all its arguments are constants; use the specific
                # call return type
                return_type = self._conditional_return_type(**crt_kwargs)

        if return_type.nullable:
            return return_type

        # If `return_type` is non-nullable, but the function call has a nullable input to any of its non-nullable
        # parameters, then we need to make it nullable. This is because Pixeltable defaults a function output to
        # `None` when any of its non-nullable inputs are `None`.
        for arg_name, arg in bound_args.items():
            param = self.signature.parameters[arg_name]
            if param.kind in (inspect.Parameter.VAR_POSITIONAL, inspect.Parameter.VAR_KEYWORD):
                continue
            if arg.col_type.nullable and not param.col_type.nullable:
                return_type = return_type.copy(nullable=True)
                break

        return return_type

    def _assemble_callable_args(
        self, callable: Callable, bound_args: dict[str, 'exprs.Expr']
    ) -> Optional[dict[str, Any]]:
        """
        Return the kwargs to pass to callable, given bound_args passed to this function.

        This is used by `conditional_return_type` and `get_resource_pool` to determine call-specific characteristics
        of this function.

        In both cases, the specified `Callable` takes a subset of the parameters of this Function, which may
        be typed as either `Expr`s or Python values. Any parameters typed as Python values expect to see constants
        (Literals); if the corresponding entries in `bound_args` are not constants, then the return value is None.
        """
        from pixeltable import exprs

        assert not self.is_polymorphic

        callable_signature = inspect.signature(callable)
        callable_type_hints = typing.get_type_hints(callable)
        callable_args: dict[str, Any] = {}

        for param in callable_signature.parameters.values():
            assert param.name in self.signature.parameters

            arg: exprs.Expr
            if param.name in bound_args:
                arg = bound_args[param.name]
            elif self.signature.parameters[param.name].has_default():
                arg = self.signature.parameters[param.name].default
            else:
                # This parameter is missing from bound_args and has no default value, so return None.
                return None
            assert isinstance(arg, exprs.Expr)

            expects_expr: Optional[type[exprs.Expr]] = None
            type_hint = callable_type_hints.get(param.name)
            if typing.get_origin(type_hint) is not None:
                type_hint = typing.get_origin(type_hint)  # Remove type subscript if one exists
            if isinstance(type_hint, type) and issubclass(type_hint, exprs.Expr):
                # The callable expects an Expr for this parameter. We allow for the case where the
                # callable requests a specific subtype of Expr.
                expects_expr = type_hint

            if expects_expr is not None:
                # The callable is expecting `param.name` to be an Expr. Validate that it's of the appropriate type;
                # otherwise return None.
                if isinstance(arg, expects_expr):
                    callable_args[param.name] = arg
                else:
                    return None
            elif isinstance(arg, exprs.Literal):
                # The callable is expecting `param.name` to be a constant Python value. Unpack a Literal if we find
                # one; otherwise return None.
                callable_args[param.name] = arg.val
            else:
                return None

        return callable_args

    def conditional_return_type(self, fn: Callable[..., ts.ColumnType]) -> Callable[..., ts.ColumnType]:
        """Instance decorator for specifying a conditional return type for this function"""
        # verify that call_return_type only has parameters that are also present in the signature
        fn_sig = inspect.signature(fn)
        for param in fn_sig.parameters.values():
            for self_sig in self.signatures:
                if param.name not in self_sig.parameters:
                    raise ValueError(
                        f'`conditional_return_type` has parameter `{param.name}` that is not in a signature'
                    )
        self._conditional_return_type = fn
        return fn

    def using(self, **kwargs: Any) -> 'ExprTemplateFunction':
        from .expr_template_function import ExprTemplateFunction

        assert len(self.signatures) > 0
        if len(self.signatures) == 1:
            # Only one signature: call _bind_and_create_template() and surface any errors directly
            template = self._bind_and_create_template(kwargs)
            return ExprTemplateFunction([template])
        else:
            # Multiple signatures: iterate over each signature and generate a template for each
            # successful binding. If there are no successful bindings, raise a generic error.
            # (Note that the resulting ExprTemplateFunction may have strictly fewer signatures than
            # this Function, in the event that only some of the signatures are successfully bound.)
            templates: list['ExprTemplate'] = []
            for i in range(len(self.signatures)):
                try:
                    template = self._resolved_fns[i]._bind_and_create_template(kwargs)
                    templates.append(template)
                except (TypeError, excs.Error):
                    continue
            if len(templates) == 0:
                raise excs.Error(f'Function {self.name!r} has no matching signature for arguments')
            return ExprTemplateFunction(templates)

    def _bind_and_create_template(self, kwargs: dict[str, Any]) -> 'ExprTemplate':
        from pixeltable import exprs

        from .expr_template_function import ExprTemplate

        assert not self.is_polymorphic

        # Resolve each kwarg into a parameter binding
        bindings: dict[str, exprs.Expr] = {}
        for k, v in kwargs.items():
            if k not in self.signature.parameters:
                raise excs.Error(f'Unknown parameter: {k}')
            param = self.signature.parameters[k]
            expr = exprs.Expr.from_object(v)
            if not isinstance(expr, exprs.Literal):
                raise excs.Error(f'Expected a constant value for parameter {k!r} in call to .using()')
            if not param.col_type.is_supertype_of(expr.col_type):
                raise excs.Error(f'Expected type `{param.col_type}` for parameter {k!r}; got `{expr.col_type}`')
            bindings[k] = expr

        residual_params = [p for p in self.signature.parameters.values() if p.name not in bindings]

        # Bind each remaining parameter to a like-named variable.
        # Also construct the call arguments for the template function call. Variables become args when possible;
        # otherwise, they are passed as kwargs.
        template_args: list[exprs.Expr] = []
        template_kwargs: dict[str, exprs.Expr] = {}
        args_ok = True
        for name, param in self.signature.parameters.items():
            if name in bindings:
                template_kwargs[name] = bindings[name]
                args_ok = False
            else:
                var = exprs.Variable(name, param.col_type)
                bindings[name] = var
                if args_ok and param.kind in (
                    inspect.Parameter.POSITIONAL_ONLY,
                    inspect.Parameter.POSITIONAL_OR_KEYWORD,
                ):
                    template_args.append(var)
                else:
                    template_kwargs[name] = var
                    args_ok = False

        return_type = self.call_return_type(bindings)
        call = exprs.FunctionCall(self, template_args, template_kwargs, return_type)

        # Construct the (n-k)-ary signature of the new function. We use `call.col_type` for this, rather than
        # `self.signature.return_type`, because the return type of the new function may be specialized via a
        # conditional return type.
        new_signature = Signature(call.col_type, residual_params, self.signature.is_batched)

        return ExprTemplate(call, new_signature)

    def exec(self, args: Sequence[Any], kwargs: dict[str, Any]) -> Any:
        """Execute the function with the given arguments and return the result."""
        raise NotImplementedError()

    async def aexec(self, *args: Any, **kwargs: Any) -> Any:
        """Execute the function with the given arguments and return the result."""
        raise NotImplementedError()

    def to_sql(self, fn: Callable[..., Optional[sql.ColumnElement]]) -> Callable[..., Optional[sql.ColumnElement]]:
        """Instance decorator for specifying the SQL translation of this function"""
        self._to_sql = fn
        return fn

    def __default_to_sql(self, *args: Any, **kwargs: Any) -> Optional[sql.ColumnElement]:
        """The default implementation of SQL translation, which provides no translation"""
        return None

    def resource_pool(self, fn: Callable[..., str]) -> Callable[..., str]:
        """Instance decorator for specifying the resource pool of this function"""
        # TODO: check that fn's parameters are a subset of our parameters
        self._resource_pool = fn
        return fn

    def __default_resource_pool(self) -> Optional[str]:
        return None

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, self.__class__):
            return False
        return self.self_path == other.self_path

    def __hash__(self) -> int:
        return hash(self.self_path)

    def source(self) -> None:
        """Print source code"""
        print('source not available')

    def as_dict(self) -> dict[str, Any]:
        """
        Return a serialized reference to the instance that can be passed to json.dumps() and converted back
        to an instance with from_dict().
        Subclasses can override _as_dict().
        """
        # We currently only ever serialize a function that has a specific signature (not a polymorphic form).
        assert not self.is_polymorphic
        classpath = f'{self.__class__.__module__}.{self.__class__.__qualname__}'
        return {'_classpath': classpath, **self._as_dict()}

    def _as_dict(self) -> dict:
        """Default serialization: store the path to self (which includes the module path) and signature."""
        assert self.self_path is not None
        return {'path': self.self_path, 'signature': self.signature.as_dict()}

    @classmethod
    def from_dict(cls, d: dict) -> Function:
        """
        Turn dict that was produced by calling as_dict() into an instance of the correct Function subclass.
        """
        assert '_classpath' in d
        module_path, class_name = d['_classpath'].rsplit('.', 1)
        class_module = importlib.import_module(module_path)
        func_class = getattr(class_module, class_name)
        assert isinstance(func_class, type) and issubclass(func_class, Function)
        return func_class._from_dict(d)

    @classmethod
    def _from_dict(cls, d: dict) -> Function:
        """Default deserialization: load the symbol indicated by the stored symbol_path"""
        path = d.get('path')
        assert path is not None
        try:
            instance = resolve_symbol(path)
            if isinstance(instance, Function):
                return instance
            else:
                return InvalidFunction(
                    path, d, f'the symbol {path!r} is no longer a UDF. (Was the `@pxt.udf` decorator removed?)'
                )
        except (AttributeError, ImportError):
            return InvalidFunction(path, d, f'the symbol {path!r} no longer exists. (Was the UDF moved or renamed?)')

    def to_store(self) -> tuple[dict, bytes]:
        """
        Serialize the function to a format that can be stored in the Pixeltable store
        Returns:
            - a dict that can be passed to json.dumps()
            - additional binary data
        Only Function subclasses that can be stored need to override this.
        """
        raise NotImplementedError()

    @classmethod
    def from_store(cls, name: Optional[str], md: dict, binary_obj: bytes) -> Function:
        """
        Create a Function instance from the serialized representation returned by to_store()
        """
        raise NotImplementedError()


class InvalidFunction(Function):
    fn_dict: dict[str, Any]
    error_msg: str

    def __init__(self, self_path: str, fn_dict: dict[str, Any], error_msg: str):
        super().__init__([], self_path)
        self.fn_dict = fn_dict
        self.error_msg = error_msg

    def _as_dict(self) -> dict:
        """
        Here we write out (verbatim) the original metadata that failed to load (and that resulted in the
        InvalidFunction). Note that the InvalidFunction itself is never serialized, so there is no corresponding
        from_dict() method.
        """
        return self.fn_dict

    @property
    def is_async(self) -> bool:
        return False



from __future__ import annotations

import abc
import hashlib
import importlib
import inspect
import json
import sys
import typing
from typing import TYPE_CHECKING, Any, Callable, Iterable, Iterator, Optional, TypeVar, Union, overload
from uuid import UUID

import numpy as np
import sqlalchemy as sql
from typing_extensions import Self, _AnnotatedAlias

from pixeltable import catalog, exceptions as excs, func, type_system as ts

from .data_row import DataRow
from .globals import ArithmeticOperator, ComparisonOperator, LiteralPythonTypes, LogicalOperator, StringOperator

if TYPE_CHECKING:
    from pixeltable import exprs


class ExprScope:
    """
    Representation of the scope in which an Expr needs to be evaluated. Used to determine nesting of scopes.
    parent is None: outermost scope
    """

    def __init__(self, parent: Optional[ExprScope]):
        self.parent = parent

    def is_contained_in(self, other: ExprScope) -> bool:
        if self == other:
            return True
        if self.parent is None:
            return False
        return self.parent.is_contained_in(other)


_GLOBAL_SCOPE = ExprScope(None)


class Expr(abc.ABC):
    """
    Rules for using state in subclasses:
    - all state except for components and slot_idx is shared between copies of an Expr
    - slot_idx is set during analysis (DataFrame.show())
    - during eval(), components can only be accessed via self.components; any Exprs outside of that won't
      have slot_idx set
    """

    col_type: ts.ColumnType

    # the subexprs are needed to construct this expr
    components: list[Expr]

    # each instance has an id that is used for equality comparisons
    # - set by the subclass's __init__()
    # - produced by _create_id()
    # - not expected to survive a serialize()/deserialize() roundtrip
    id: Optional[int]

    # index of the expr's value in the data row:
    # - set for all materialized exprs
    # - None: not executable
    # - not set for subexprs that don't need to be materialized because the parent can be materialized via SQL
    slot_idx: Optional[int]

    T = TypeVar('T', bound='Expr')

    def __init__(self, col_type: ts.ColumnType):
        self.col_type = col_type
        self.components = []
        self.id = None
        self.slot_idx = None

    def dependencies(self) -> list[Expr]:
        """
        Returns all exprs that need to have been evaluated before eval() can be called on this one.
        """
        return self.components

    def scope(self) -> ExprScope:
        # by default this is the innermost scope of any of our components
        result = _GLOBAL_SCOPE
        for c in self.components:
            c_scope = c.scope()
            if c_scope.is_contained_in(result):
                result = c_scope
        return result

    def bind_rel_paths(self) -> None:
        """
        Binds relative JsonPaths to mapper.
        This needs to be done in a separate phase after __init__(), because RelativeJsonPath()(-1) cannot be resolved
        by the immediately containing JsonMapper during initialization.
        """
        self._bind_rel_paths()
        has_rel_path = self._has_relative_path()
        assert not has_rel_path, self._expr_tree()
        assert not self._has_relative_path(), self._expr_tree()

    def _bind_rel_paths(self, mapper: Optional['exprs.JsonMapperDispatch'] = None) -> None:
        for c in self.components:
            c._bind_rel_paths(mapper)

    def _expr_tree(self) -> str:
        """Returns a string representation of this expression as a multi-line tree. Useful for debugging."""
        buf: list[str] = []
        self._expr_tree_r(0, buf)
        return '\n'.join(buf)

    def _expr_tree_r(self, indent: int, buf: list[str]) -> None:
        buf.append(f'{" " * indent}{type(self).__name__}: {self}'.replace('\n', '\\n'))
        for c in self.components:
            c._expr_tree_r(indent + 2, buf)

    def default_column_name(self) -> Optional[str]:
        """
        Returns:
            None if this expression lacks a default name,
            or a valid identifier (according to catalog.is_valid_identifer) otherwise.
        """
        return None

    @property
    def validation_error(self) -> Optional[str]:
        """
        Subclasses can override this to indicate that validation has failed after a catalog load.

        If an Expr (or any of its transitive components) is invalid, then it cannot be evaluated, but its metadata
        will still be preserved in the catalog (so that the user can take appropriate corrective action).
        """
        for c in self.components:
            error = c.validation_error
            if error is not None:
                return error
        return None

    @property
    def is_valid(self) -> bool:
        return self.validation_error is None

    def equals(self, other: Expr) -> bool:
        """
        Subclass-specific comparison. Implemented as a function because __eq__() is needed to construct Comparisons.
        """
        if type(self) is not type(other):
            return False
        if len(self.components) != len(other.components):
            return False
        for i in range(len(self.components)):
            if not self.components[i].equals(other.components[i]):
                return False
        return self._equals(other)

    def _equals(self, other: Self) -> bool:
        # we already compared the type and components in equals(); subclasses that require additional comparisons
        # override this
        return True

    def _id_attrs(self) -> list[tuple[str, Any]]:
        """Returns attribute name/value pairs that are used to construct the instance id.

        Attribute values must be immutable and have str() defined.
        """
        return [('classname', self.__class__.__name__)]

    def _create_id(self) -> int:
        hasher = hashlib.sha256()
        for attr, value in self._id_attrs():
            hasher.update(attr.encode('utf-8'))
            hasher.update(str(value).encode('utf-8'))
        # Include the col_type of the expression to avoid expressions with identical str() representations
        # but different types being considered the same expression, e.g. str(int(4)) == "4"
        hasher.update(repr(self.col_type).encode('utf-8'))
        for expr in self.components:
            hasher.update(str(expr.id).encode('utf-8'))
        # truncate to machine's word size
        return int(hasher.hexdigest(), 16) & sys.maxsize

    def __hash__(self) -> int:
        assert self.id is not None
        return self.id

    @classmethod
    def list_equals(cls, a: list[Expr], b: list[Expr]) -> bool:
        if len(a) != len(b):
            return False
        return all(a[i].equals(b[i]) for i in range(len(a)))

    def copy(self: T) -> T:
        """
        Creates a copy that can be evaluated separately: it doesn't share any eval context (slot_idx)
        but shares everything else (catalog objects, etc.)
        """
        cls = self.__class__
        result = cls.__new__(cls)
        result.__dict__.update(self.__dict__)
        result.slot_idx = None
        result.components = [c.copy() for c in self.components]
        return result

    @classmethod
    def copy_list(cls, expr_list: Optional[list[Expr]]) -> Optional[list[Expr]]:
        if expr_list is None:
            return None
        return [e.copy() for e in expr_list]

    def __deepcopy__(self, memo: Optional[dict[int, Any]] = None) -> Expr:
        # we don't need to create an actual deep copy because all state other than execution state is read-only
        if memo is None:
            memo = {}
        result = self.copy()
        memo[id(self)] = result
        return result

    def substitute(self, spec: dict[Expr, Expr]) -> Expr:
        """
        Replace 'old' with 'new' recursively, and return a new version of the expression
        This method must be used in the form: expr = expr.substitute(spec)
        """
        from .literal import Literal

        if isinstance(self, Literal):
            return self
        for old, new in spec.items():
            if self.equals(old):
                return new.copy()
        for i in range(len(self.components)):
            self.components[i] = self.components[i].substitute(spec)
        result = self.maybe_literal()
        result.id = result._create_id()
        return result

    @classmethod
    def list_substitute(cls, expr_list: list[Expr], spec: dict[Expr, Expr]) -> None:
        for i in range(len(expr_list)):
            expr_list[i] = expr_list[i].substitute(spec)

    def resolve_computed_cols(self, resolve_cols: Optional[set[catalog.Column]] = None) -> Expr:
        """
        Recursively replace ColRefs to unstored computed columns with their value exprs.
        Also replaces references to stored computed columns in resolve_cols.
        """
        from .column_ref import ColumnRef
        from .expr_set import ExprSet

        if resolve_cols is None:
            resolve_cols = set()
        result = self
        while True:
            target_col_refs = ExprSet(
                [
                    e
                    for e in result.subexprs()
                    if isinstance(e, ColumnRef) and e.col.is_computed and (not e.col.is_stored or e.col in resolve_cols)
                ]
            )
            if len(target_col_refs) == 0:
                return result
            result = result.substitute({ref: ref.col.value_expr for ref in target_col_refs})

    def is_bound_by(self, tbls: list[catalog.TableVersionPath]) -> bool:
        """Returns True if this expr can be evaluated in the context of tbls."""
        from .column_ref import ColumnRef

        col_refs = self.subexprs(ColumnRef)
        return all(any(tbl.has_column(col_ref.col) for tbl in tbls) for col_ref in col_refs)

    def retarget(self, tbl: catalog.TableVersionPath) -> Self:
        """Retarget ColumnRefs in this expr to the specific TableVersions in tbl."""
        tbl_versions = {tbl_version.id: tbl_version.get() for tbl_version in tbl.get_tbl_versions()}
        return self._retarget(tbl_versions)

    @classmethod
    def retarget_list(cls, expr_list: list[Expr], tbl: catalog.TableVersionPath) -> None:
        """Retarget ColumnRefs in expr_list to the specific TableVersions in tbl."""
        tbl_versions = {tbl_version.id: tbl_version.get() for tbl_version in tbl.get_tbl_versions()}
        for i in range(len(expr_list)):
            expr_list[i] = expr_list[i]._retarget(tbl_versions)

    def _retarget(self, tbl_versions: dict[UUID, catalog.TableVersion]) -> Self:
        for i in range(len(self.components)):
            self.components[i] = self.components[i]._retarget(tbl_versions)
        return self

    def __repr__(self) -> str:
        return f'<Expression of type {type(self)}>'

    def display_str(self, inline: bool = True) -> str:
        """
        inline: if False, use line breaks where appropriate; otherwise don't use linebreaks
        """
        return str(self)

    @classmethod
    def print_list(cls, expr_list: list[Any]) -> str:
        if len(expr_list) == 1:
            return str(expr_list[0])
        return f'({", ".join(str(e) for e in expr_list)})'

    # `subexprs` has two forms: one that takes an explicit subclass of `Expr` as an argument and returns only
    # instances of that subclass; and another that returns all subexpressions that match the given filter.
    # In order for type checking to behave correctly on both forms, we provide two overloaded signatures.

    @overload
    def subexprs(
        self, *, filter: Optional[Callable[[Expr], bool]] = None, traverse_matches: bool = True
    ) -> Iterator[Expr]: ...

    @overload
    def subexprs(
        self, expr_class: type[T], filter: Optional[Callable[[Expr], bool]] = None, traverse_matches: bool = True
    ) -> Iterator[T]: ...

    def subexprs(
        self,
        expr_class: Optional[type[T]] = None,
        filter: Optional[Callable[[Expr], bool]] = None,
        traverse_matches: bool = True,
    ) -> Iterator[T]:
        """
        Iterate over all subexprs, including self.
        """
        is_match = isinstance(self, expr_class) if expr_class is not None else True
        # apply filter after checking for expr_class
        if filter is not None and is_match:
            is_match = filter(self)
        if not is_match or traverse_matches:
            for c in self.components:
                yield from c.subexprs(expr_class=expr_class, filter=filter, traverse_matches=traverse_matches)
        if is_match:
            yield self  # type: ignore[misc]

    @overload
    @classmethod
    def list_subexprs(
        cls,
        expr_list: Iterable[Expr],
        *,
        filter: Optional[Callable[[Expr], bool]] = None,
        traverse_matches: bool = True,
    ) -> Iterator[Expr]: ...

    @overload
    @classmethod
    def list_subexprs(
        cls,
        expr_list: Iterable[Expr],
        expr_class: type[T],
        filter: Optional[Callable[[Expr], bool]] = None,
        traverse_matches: bool = True,
    ) -> Iterator[T]: ...

    @classmethod
    def list_subexprs(
        cls,
        expr_list: Iterable[Expr],
        expr_class: Optional[type[T]] = None,
        filter: Optional[Callable[[Expr], bool]] = None,
        traverse_matches: bool = True,
    ) -> Iterator[T]:
        """Produce subexprs for all exprs in list. Can contain duplicates."""
        for e in expr_list:
            yield from e.subexprs(expr_class=expr_class, filter=filter, traverse_matches=traverse_matches)

    def _contains(self, cls: Optional[type[Expr]] = None, filter: Optional[Callable[[Expr], bool]] = None) -> bool:
        """
        Returns True if any subexpr is an instance of cls and/or matches filter.
        """
        assert cls is not None or filter is not None
        try:
            _ = next(self.subexprs(expr_class=cls, filter=filter, traverse_matches=False))
            return True
        except StopIteration:
            return False

    def _has_relative_path(self) -> bool:
        return any(c._has_relative_path() for c in self.components)

    def tbl_ids(self) -> set[UUID]:
        """Returns table ids referenced by this expr."""
        from .column_ref import ColumnRef
        from .rowid_ref import RowidRef

        return {ref.col.tbl.id for ref in self.subexprs(ColumnRef)} | {ref.tbl.id for ref in self.subexprs(RowidRef)}

    @classmethod
    def all_tbl_ids(cls, exprs_: Iterable[Expr]) -> set[UUID]:
        return {tbl_id for e in exprs_ for tbl_id in e.tbl_ids()}

    @classmethod
    def get_refd_column_ids(cls, expr_dict: dict[str, Any]) -> set[catalog.QColumnId]:
        """Return Columns referenced by expr_dict."""
        result: set[catalog.QColumnId] = set()
        assert '_classname' in expr_dict
        from .column_ref import ColumnRef

        if expr_dict['_classname'] == 'ColumnRef':
            result.add(ColumnRef.get_column_id(expr_dict))
        if 'components' in expr_dict:
            for component_dict in expr_dict['components']:
                result.update(cls.get_refd_column_ids(component_dict))
        return result

    def as_literal(self) -> Optional[Expr]:
        """
        Return a Literal expression if this expression can be evaluated to a constant value, otherwise return None.
        """
        return None

    @classmethod
    def from_array(cls, elements: Iterable) -> Optional[Expr]:
        from .inline_expr import InlineArray
        from .literal import Literal

        if isinstance(elements, np.ndarray):
            pxttype = ts.ArrayType.from_literal(elements)
            if pxttype is not None:
                return Literal(elements, col_type=pxttype)

        inline_array = InlineArray(elements)
        return inline_array.maybe_literal()

    def maybe_literal(self: Expr) -> Expr:
        """
        Return a Literal if this expression can be evaluated to a constant value, otherwise return the expression.
        """
        lit_expr = self.as_literal()
        if lit_expr is not None:
            return lit_expr
        else:
            return self

    @classmethod
    def from_object(cls, o: object) -> Optional[Expr]:
        """
        Try to turn a literal object into an Expr.
        """
        from .inline_expr import InlineDict, InlineList
        from .literal import Literal

        # Try to create a literal. We need to check for InlineList/InlineDict
        # first, to prevent them from inappropriately being interpreted as JsonType
        # literals.
        if isinstance(o, Literal):
            return o

        if isinstance(o, (list, tuple, dict, Expr)):
            expr: Expr
            if isinstance(o, (list, tuple)):
                expr = InlineList(o)
            elif isinstance(o, dict):
                expr = InlineDict(o)
            else:
                expr = o

            return expr.maybe_literal()
        else:
            # convert scalar to a literal
            obj_type = ts.ColumnType.infer_literal_type(o)
            if obj_type is not None:
                return Literal(o, col_type=obj_type)
        return None

    def sql_expr(self, sql_elements: 'exprs.SqlElementCache') -> Optional[sql.ColumnElement]:
        """
        If this expr can be materialized directly in SQL:
        - returns a ColumnElement
        - eval() will not be called (exception: Literal)
        Otherwise
        - returns None
        - eval() will be called
        """
        return None

    @abc.abstractmethod
    def eval(self, data_row: DataRow, row_builder: 'exprs.RowBuilder') -> None:
        """
        Compute the expr value for data_row and store the result in data_row[slot_idx].
        Not called if sql_expr() != None (exception: Literal).
        """
        pass

    def release(self) -> None:
        """
        Allow Expr class to tear down execution state. This is called after the last eval() call.
        """
        for c in self.components:
            c.release()

    @classmethod
    def release_list(cls, expr_list: list[Expr]) -> None:
        for e in expr_list:
            e.release()

    def serialize(self) -> str:
        return json.dumps(self.as_dict())

    def as_dict(self) -> dict:
        """
        Turn Expr object into a dict that can be passed to json.dumps().
        Subclasses override _as_dict().
        """
        return {'_classname': self.__class__.__name__, **self._as_dict()}

    @classmethod
    def as_dict_list(cls, expr_list: list[Expr]) -> list[dict]:
        return [e.as_dict() for e in expr_list]

    def _as_dict(self) -> dict:
        if len(self.components) > 0:
            return {'components': [c.as_dict() for c in self.components]}
        return {}

    @classmethod
    def deserialize(cls, dict_str: str) -> Expr:
        return cls.from_dict(json.loads(dict_str))

    @classmethod
    def from_dict(cls, d: dict) -> Self:
        """
        Turn dict that was produced by calling Expr.as_dict() into an instance of the correct Expr subclass.
        """
        assert '_classname' in d
        exprs_module = importlib.import_module(cls.__module__.rsplit('.', 1)[0])
        type_class = getattr(exprs_module, d['_classname'])
        components: list[Expr] = []
        if 'components' in d:
            components = [cls.from_dict(component_dict) for component_dict in d['components']]
        return type_class._from_dict(d, components)

    @classmethod
    def from_dict_list(cls, dict_list: list[dict]) -> list[Expr]:
        return [cls.from_dict(d) for d in dict_list]

    @classmethod
    def _from_dict(cls, d: dict, components: list[Expr]) -> Self:
        raise AssertionError(f'not implemented: {cls.__name__}')

    def isin(self, value_set: Any) -> 'exprs.InPredicate':
        from .in_predicate import InPredicate

        if isinstance(value_set, Expr):
            return InPredicate(self, value_set_expr=value_set)
        else:
            return InPredicate(self, value_set_literal=value_set)

    def astype(self, new_type: Union[ts.ColumnType, type, _AnnotatedAlias]) -> 'exprs.TypeCast':
        from pixeltable.exprs import TypeCast

        # Interpret the type argument the same way we would if given in a schema
        col_type = ts.ColumnType.normalize_type(new_type, nullable_default=True, allow_builtin_types=False)
        if not self.col_type.nullable:
            # This expression is non-nullable; we can prove that the output is non-nullable, regardless of
            # whether new_type is given as nullable.
            col_type = col_type.copy(nullable=False)
        return TypeCast(self, col_type)

    def apply(
        self, fn: Callable, *, col_type: Union[ts.ColumnType, type, _AnnotatedAlias, None] = None
    ) -> 'exprs.FunctionCall':
        if col_type is not None:
            col_type = ts.ColumnType.normalize_type(col_type)
        function = self._make_applicator_function(fn, col_type)
        # Return a `FunctionCall` obtained by passing this `Expr` to the new `function`.
        return function(self)

    def __dir__(self) -> list[str]:
        attrs = ['isin', 'astype', 'apply']
        attrs += [f.name for f in func.FunctionRegistry.get().get_type_methods(self.col_type.type_enum)]
        return attrs

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        raise NotImplementedError(f'Expression of type `{type(self)}` is not callable')

    def __getitem__(self, index: object) -> Expr:
        if self.col_type.is_json_type():
            from .json_path import JsonPath

            return JsonPath(self)[index]
        if self.col_type.is_array_type():
            from .array_slice import ArraySlice

            if not isinstance(index, tuple):
                index = (index,)
            if any(not isinstance(i, (int, slice)) for i in index):
                raise AttributeError(f'Invalid array indices: {index}')
            return ArraySlice(self, index)
        raise AttributeError(f'Type {self.col_type} is not subscriptable')

    def __getattr__(self, name: str) -> 'exprs.Expr':
        """
        ex.: <img col>.rotate(60)
        """
        from .json_path import JsonPath
        from .method_ref import MethodRef

        if self.col_type.is_json_type():
            return JsonPath(self).__getattr__(name)
        else:
            method_ref = MethodRef(self, name)
            if method_ref.fn.is_property:
                # Marked as a property, so autoinvoke the method to obtain a `FunctionCall`
                assert method_ref.fn.arity == 1
                return method_ref.fn(method_ref.base_expr)
            else:
                # Return the `MethodRef` object itself; it requires arguments to become a `FunctionCall`
                return method_ref

    def __bool__(self) -> bool:
        raise TypeError(
            f'Pixeltable expressions cannot be used in conjunction with Python boolean operators (and/or/not)\n{self!r}'
        )

    def __lt__(self, other: object) -> 'exprs.Comparison':
        return self._make_comparison(ComparisonOperator.LT, other)

    def __le__(self, other: object) -> 'exprs.Comparison':
        return self._make_comparison(ComparisonOperator.LE, other)

    def __eq__(self, other: object) -> 'exprs.Expr':  # type: ignore[override]
        if other is None:
            from .is_null import IsNull

            return IsNull(self)
        return self._make_comparison(ComparisonOperator.EQ, other)

    def __ne__(self, other: object) -> 'exprs.Expr':  # type: ignore[override]
        if other is None:
            from .compound_predicate import CompoundPredicate
            from .is_null import IsNull

            return CompoundPredicate(LogicalOperator.NOT, [IsNull(self)])
        return self._make_comparison(ComparisonOperator.NE, other)

    def __gt__(self, other: object) -> 'exprs.Comparison':
        return self._make_comparison(ComparisonOperator.GT, other)

    def __ge__(self, other: object) -> 'exprs.Comparison':
        return self._make_comparison(ComparisonOperator.GE, other)

    def _make_comparison(self, op: ComparisonOperator, other: object) -> 'exprs.Comparison':
        """
        other: Union[Expr, LiteralPythonTypes]
        """
        # TODO: check for compatibility
        from .comparison import Comparison
        from .literal import Literal

        if isinstance(other, Expr):
            return Comparison(op, self, other)
        if isinstance(other, typing.get_args(LiteralPythonTypes)):
            return Comparison(op, self, Literal(other))
        raise TypeError(f'Other must be Expr or literal: {type(other)}')

    def __neg__(self) -> 'exprs.ArithmeticExpr':
        return self._make_arithmetic_expr(ArithmeticOperator.MUL, -1)

    def __add__(self, other: object) -> Union[exprs.ArithmeticExpr, exprs.StringOp]:
        if isinstance(self, str) or (isinstance(self, Expr) and self.col_type.is_string_type()):
            return self._make_string_expr(StringOperator.CONCAT, other)
        return self._make_arithmetic_expr(ArithmeticOperator.ADD, other)

    def __sub__(self, other: object) -> 'exprs.ArithmeticExpr':
        return self._make_arithmetic_expr(ArithmeticOperator.SUB, other)

    def __mul__(self, other: object) -> Union['exprs.ArithmeticExpr', 'exprs.StringOp']:
        if isinstance(self, str) or (isinstance(self, Expr) and self.col_type.is_string_type()):
            return self._make_string_expr(StringOperator.REPEAT, other)
        return self._make_arithmetic_expr(ArithmeticOperator.MUL, other)

    def __truediv__(self, other: object) -> 'exprs.ArithmeticExpr':
        return self._make_arithmetic_expr(ArithmeticOperator.DIV, other)

    def __mod__(self, other: object) -> 'exprs.ArithmeticExpr':
        return self._make_arithmetic_expr(ArithmeticOperator.MOD, other)

    def __floordiv__(self, other: object) -> 'exprs.ArithmeticExpr':
        return self._make_arithmetic_expr(ArithmeticOperator.FLOORDIV, other)

    def __radd__(self, other: object) -> Union['exprs.ArithmeticExpr', 'exprs.StringOp']:
        if isinstance(other, str) or (isinstance(other, Expr) and other.col_type.is_string_type()):
            return self._rmake_string_expr(StringOperator.CONCAT, other)
        return self._rmake_arithmetic_expr(ArithmeticOperator.ADD, other)

    def __rsub__(self, other: object) -> 'exprs.ArithmeticExpr':
        return self._rmake_arithmetic_expr(ArithmeticOperator.SUB, other)

    def __rmul__(self, other: object) -> Union['exprs.ArithmeticExpr', 'exprs.StringOp']:
        if isinstance(other, str) or (isinstance(other, Expr) and other.col_type.is_string_type()):
            return self._rmake_string_expr(StringOperator.REPEAT, other)
        return self._rmake_arithmetic_expr(ArithmeticOperator.MUL, other)

    def __rtruediv__(self, other: object) -> 'exprs.ArithmeticExpr':
        return self._rmake_arithmetic_expr(ArithmeticOperator.DIV, other)

    def __rmod__(self, other: object) -> 'exprs.ArithmeticExpr':
        return self._rmake_arithmetic_expr(ArithmeticOperator.MOD, other)

    def __rfloordiv__(self, other: object) -> 'exprs.ArithmeticExpr':
        return self._rmake_arithmetic_expr(ArithmeticOperator.FLOORDIV, other)

    def _make_string_expr(self, op: StringOperator, other: object) -> 'exprs.StringOp':
        """
        Make left-handed version of string expression.
        """
        from .literal import Literal
        from .string_op import StringOp

        if isinstance(other, Expr):
            return StringOp(op, self, other)
        if isinstance(other, typing.get_args(LiteralPythonTypes)):
            return StringOp(op, self, Literal(other))
        raise TypeError(f'Other must be Expr or literal: {type(other)}')

    def _rmake_string_expr(self, op: StringOperator, other: object) -> 'exprs.StringOp':
        """
        Right-handed version of _make_string_expr. other must be a literal; if it were an Expr,
        the operation would have already been evaluated in its left-handed form.
        """
        from .literal import Literal
        from .string_op import StringOp

        assert not isinstance(other, Expr)  # Else the left-handed form would have evaluated first
        if isinstance(other, typing.get_args(LiteralPythonTypes)):
            return StringOp(op, Literal(other), self)
        raise TypeError(f'Other must be Expr or literal: {type(other)}')

    def _make_arithmetic_expr(self, op: ArithmeticOperator, other: object) -> 'exprs.ArithmeticExpr':
        """
        other: Union[Expr, LiteralPythonTypes]
        """
        # TODO: check for compatibility
        from .arithmetic_expr import ArithmeticExpr
        from .literal import Literal

        if isinstance(other, Expr):
            return ArithmeticExpr(op, self, other)
        if isinstance(other, typing.get_args(LiteralPythonTypes)):
            return ArithmeticExpr(op, self, Literal(other))
        raise TypeError(f'Other must be Expr or literal: {type(other)}')

    def _rmake_arithmetic_expr(self, op: ArithmeticOperator, other: object) -> 'exprs.ArithmeticExpr':
        """
        Right-handed version of _make_arithmetic_expr. other must be a literal; if it were an Expr,
        the operation would have already been evaluated in its left-handed form.
        """
        # TODO: check for compatibility
        from .arithmetic_expr import ArithmeticExpr
        from .literal import Literal

        assert not isinstance(other, Expr)  # Else the left-handed form would have evaluated first
        if isinstance(other, typing.get_args(LiteralPythonTypes)):
            return ArithmeticExpr(op, Literal(other), self)
        raise TypeError(f'Other must be Expr or literal: {type(other)}')

    def __and__(self, other: object) -> Expr:
        if not isinstance(other, Expr):
            raise TypeError(f'Other needs to be an expression: {type(other)}')
        if not other.col_type.is_bool_type():
            raise TypeError(f'Other needs to be an expression that returns a boolean: {other.col_type}')
        from .compound_predicate import CompoundPredicate

        return CompoundPredicate(LogicalOperator.AND, [self, other])

    def __or__(self, other: object) -> Expr:
        if not isinstance(other, Expr):
            raise TypeError(f'Other needs to be an expression: {type(other)}')
        if not other.col_type.is_bool_type():
            raise TypeError(f'Other needs to be an expression that returns a boolean: {other.col_type}')
        from .compound_predicate import CompoundPredicate

        return CompoundPredicate(LogicalOperator.OR, [self, other])

    def __invert__(self) -> Expr:
        from .compound_predicate import CompoundPredicate

        return CompoundPredicate(LogicalOperator.NOT, [self])

    def split_conjuncts(self, condition: Callable[[Expr], bool]) -> tuple[list[Expr], Optional[Expr]]:
        """
        Returns clauses of a conjunction that meet condition in the first element.
        The second element contains remaining clauses, rolled into a conjunction.
        """
        assert self.col_type.is_bool_type()  # only valid for predicates
        if condition(self):
            return [self], None
        else:
            return [], self

    def _make_applicator_function(self, fn: Callable, col_type: Optional[ts.ColumnType]) -> 'func.Function':
        """
        Creates a unary pixeltable `Function` that encapsulates a python `Callable`. The result type of
        the new `Function` is given by `col_type`, and its parameter type will be `self.col_type`.

        Args:
            fn: The `Callable` to encapsulate. Must have at least one parameter, and at most one required
                parameter.
            col_type: The pixeltable result type of the new `Function`.
        """
        if col_type is not None:
            # col_type is specified explicitly
            fn_type = col_type
        elif fn in _known_applicator_types:
            # For convenience, various built-ins and other Python functions that don't
            # have type hints are hardcoded
            fn_type = _known_applicator_types[fn]
        elif 'return' in typing.get_type_hints(fn):
            # Attempt to infer the column type from the return type of the callable;
            # this will set fn_type to None if it cannot be inferred
            return_type = typing.get_type_hints(fn)['return']
            fn_type = ts.ColumnType.from_python_type(return_type)
        else:
            # No type hint
            fn_type = None

        if fn_type is None:
            raise excs.Error(
                f'Column type of `{fn.__name__}` cannot be inferred. '
                f'Use `.apply({fn.__name__}, col_type=...)` to specify.'
            )

        # TODO(aaron-siegel) Currently we assume that `fn` has exactly one required parameter
        # and all optional parameters take their default values. Should we provide a more
        # flexible API? For example, by defining
        # expr.apply(fn, my_kw=my_arg)
        # to mean: transform each x by calling
        # fn(x, my_kw=my_arg)
        # In the current implementation, a lambda is needed in order to specify this pattern:
        # expr.apply(lambda x: fn(x, my_kw=my_arg))

        try:
            # If `fn` is not a builtin, we can do some basic validation to ensure it's
            # compatible with `apply`.
            params = inspect.signature(fn).parameters
            params_iter = iter(params.values())
            first_param = next(params_iter) if len(params) >= 1 else None
            second_param = next(params_iter) if len(params) >= 2 else None
            # Check that fn has at least one positional parameter
            if len(params) == 0 or first_param.kind in (inspect.Parameter.KEYWORD_ONLY, inspect.Parameter.VAR_KEYWORD):
                raise excs.Error(f'Function `{fn.__name__}` has no positional parameters.')
            # Check that fn has at most one required parameter, i.e., its second parameter
            # has no default and is not a varargs
            if (
                len(params) >= 2
                and second_param.kind not in (inspect.Parameter.VAR_POSITIONAL, inspect.Parameter.VAR_KEYWORD)
                and second_param.default is inspect.Parameter.empty
            ):
                raise excs.Error(f'Function `{fn.__name__}` has multiple required parameters.')
        except ValueError:
            # inspect.signature(fn) will raise a `ValueError` if `fn` is a builtin; I don't
            # know of any way to get the signature of a builtin, nor to check for this in
            # advance (without the try/except pattern). For now, builtins will not be
            # validated.
            pass

        # Since `fn` might have optional parameters, we wrap it in a lambda to get a unary
        # equivalent, so that its signature is understood by `make_function`. This also
        # ensures that `decorated_fn` is never a builtin.
        # We also set the display_name explicitly, so that the `FunctionCall` gets the
        # name of `decorated_fn`, not the lambda.
        return func.make_function(
            decorated_fn=lambda x: fn(x), return_type=fn_type, param_types=[self.col_type], function_name=fn.__name__
        )


# A dictionary of result types of various stdlib functions that are
# commonly used in computed columns. stdlib does not have type hints, so these
# are used to infer their result types (as pixeltable types) to avoid having
# to specify them explicitly in Expr.apply().
# This is purely for convenience and does not impact the supported functionality
# (it's always possible to specify a result type explicitly for a function
# that does not have type hints and is not present in this dict).
_known_applicator_types: dict[Callable, ts.ColumnType] = {
    str: ts.StringType(),
    json.dumps: ts.StringType(),
    json.loads: ts.JsonType(),
}



"""
Pixeltable [UDFs](https://pixeltable.readme.io/docs/user-defined-functions-udfs)
that wrap various models from the Hugging Face `transformers` package.

These UDFs will cause Pixeltable to invoke the relevant models locally. In order to use them, you must
first `pip install transformers` (or in some cases, `sentence-transformers`, as noted in the specific
UDFs).
"""

from typing import Any, Callable, Optional, TypeVar

import PIL.Image

import pixeltable as pxt
import pixeltable.exceptions as excs
import pixeltable.type_system as ts
from pixeltable import env
from pixeltable.func import Batch
from pixeltable.functions.util import normalize_image_mode, resolve_torch_device
from pixeltable.utils.code import local_public_names


@pxt.udf(batch_size=32)
def sentence_transformer(
    sentence: Batch[str], *, model_id: str, normalize_embeddings: bool = False
) -> Batch[pxt.Array[(None,), pxt.Float]]:
    """
    Computes sentence embeddings. `model_id` should be a pretrained Sentence Transformers model, as described
    in the [Sentence Transformers Pretrained Models](https://sbert.net/docs/sentence_transformer/pretrained_models.html)
    documentation.

    __Requirements:__

    - `pip install torch sentence-transformers`

    Args:
        sentence: The sentence to embed.
        model_id: The pretrained model to use for the encoding.
        normalize_embeddings: If `True`, normalizes embeddings to length 1; see the
            [Sentence Transformers API Docs](https://sbert.net/docs/package_reference/sentence_transformer/SentenceTransformer.html)
            for more details

    Returns:
        An array containing the output of the embedding model.

    Examples:
        Add a computed column that applies the model `all-mpnet-base-2` to an existing Pixeltable column `tbl.sentence`
        of the table `tbl`:

        >>> tbl.add_computed_column(result=sentence_transformer(tbl.sentence, model_id='all-mpnet-base-v2'))
    """
    env.Env.get().require_package('sentence_transformers')
    device = resolve_torch_device('auto')
    from sentence_transformers import SentenceTransformer

    # specifying the device, moves the model to device (gpu:cuda/mps, cpu)
    model = _lookup_model(model_id, SentenceTransformer, device=device, pass_device_to_create=True)

    # specifying the device, uses it for computation
    array = model.encode(sentence, device=device, normalize_embeddings=normalize_embeddings)
    return [array[i] for i in range(array.shape[0])]


@sentence_transformer.conditional_return_type
def _(model_id: str) -> ts.ArrayType:
    try:
        from sentence_transformers import SentenceTransformer

        model = _lookup_model(model_id, SentenceTransformer)
        return ts.ArrayType((model.get_sentence_embedding_dimension(),), dtype=ts.FloatType(), nullable=False)
    except ImportError:
        return ts.ArrayType((None,), dtype=ts.FloatType(), nullable=False)


@pxt.udf
def sentence_transformer_list(sentences: list, *, model_id: str, normalize_embeddings: bool = False) -> list:
    env.Env.get().require_package('sentence_transformers')
    device = resolve_torch_device('auto')
    from sentence_transformers import SentenceTransformer

    # specifying the device, moves the model to device (gpu:cuda/mps, cpu)
    model = _lookup_model(model_id, SentenceTransformer, device=device, pass_device_to_create=True)

    # specifying the device, uses it for computation
    array = model.encode(sentences, device=device, normalize_embeddings=normalize_embeddings)
    return [array[i].tolist() for i in range(array.shape[0])]


@pxt.udf(batch_size=32)
def cross_encoder(sentences1: Batch[str], sentences2: Batch[str], *, model_id: str) -> Batch[float]:
    """
    Performs predicts on the given sentence pair.
    `model_id` should be a pretrained Cross-Encoder model, as described in the
    [Cross-Encoder Pretrained Models](https://www.sbert.net/docs/cross_encoder/pretrained_models.html)
    documentation.

    __Requirements:__

    - `pip install torch sentence-transformers`

    Parameters:
        sentences1: The first sentence to be paired.
        sentences2: The second sentence to be paired.
        model_id: The identifier of the cross-encoder model to use.

    Returns:
        The similarity score between the inputs.

    Examples:
        Add a computed column that applies the model `ms-marco-MiniLM-L-4-v2` to the sentences in
        columns `tbl.sentence1` and `tbl.sentence2`:

        >>> tbl.add_computed_column(result=sentence_transformer(
        ...     tbl.sentence1, tbl.sentence2, model_id='ms-marco-MiniLM-L-4-v2'
        ... ))
    """
    env.Env.get().require_package('sentence_transformers')
    device = resolve_torch_device('auto')
    from sentence_transformers import CrossEncoder

    # specifying the device, moves the model to device (gpu:cuda/mps, cpu)
    # and uses the device for predict computation
    model = _lookup_model(model_id, CrossEncoder, device=device, pass_device_to_create=True)

    array = model.predict([[s1, s2] for s1, s2 in zip(sentences1, sentences2)], convert_to_numpy=True)
    return array.tolist()


@pxt.udf
def cross_encoder_list(sentence1: str, sentences2: list, *, model_id: str) -> list:
    env.Env.get().require_package('sentence_transformers')
    device = resolve_torch_device('auto')
    from sentence_transformers import CrossEncoder

    # specifying the device, moves the model to device (gpu:cuda/mps, cpu)
    # and uses the device for predict computation
    model = _lookup_model(model_id, CrossEncoder, device=device, pass_device_to_create=True)

    array = model.predict([[sentence1, s2] for s2 in sentences2], convert_to_numpy=True)
    return array.tolist()


@pxt.udf(batch_size=32)
def clip(text: Batch[str], *, model_id: str) -> Batch[pxt.Array[(None,), pxt.Float]]:
    """
    Computes a CLIP embedding for the specified text or image. `model_id` should be a reference to a pretrained
    [CLIP Model](https://huggingface.co/docs/transformers/model_doc/clip).

    __Requirements:__

    - `pip install torch transformers`

    Args:
        text: The string to embed.
        model_id: The pretrained model to use for the embedding.

    Returns:
        An array containing the output of the embedding model.

    Examples:
        Add a computed column that applies the model `openai/clip-vit-base-patch32` to an existing
        Pixeltable column `tbl.text` of the table `tbl`:

        >>> tbl.add_computed_column(
        ...     result=clip(tbl.text, model_id='openai/clip-vit-base-patch32')
        ... )

        The same would work with an image column `tbl.image` in place of `tbl.text`.
    """
    env.Env.get().require_package('transformers')
    device = resolve_torch_device('auto')
    import torch
    from transformers import CLIPModel, CLIPProcessor

    model = _lookup_model(model_id, CLIPModel.from_pretrained, device=device)
    processor = _lookup_processor(model_id, CLIPProcessor.from_pretrained)

    with torch.no_grad():
        inputs = processor(text=text, return_tensors='pt', padding=True, truncation=True)
        embeddings = model.get_text_features(**inputs.to(device)).detach().to('cpu').numpy()

    return [embeddings[i] for i in range(embeddings.shape[0])]


@clip.overload
def _(image: Batch[PIL.Image.Image], *, model_id: str) -> Batch[pxt.Array[(None,), pxt.Float]]:
    env.Env.get().require_package('transformers')
    device = resolve_torch_device('auto')
    import torch
    from transformers import CLIPModel, CLIPProcessor

    model = _lookup_model(model_id, CLIPModel.from_pretrained, device=device)
    processor = _lookup_processor(model_id, CLIPProcessor.from_pretrained)

    with torch.no_grad():
        inputs = processor(images=image, return_tensors='pt', padding=True)
        embeddings = model.get_image_features(**inputs.to(device)).detach().to('cpu').numpy()

    return [embeddings[i] for i in range(embeddings.shape[0])]


@clip.conditional_return_type
def _(model_id: str) -> ts.ArrayType:
    try:
        from transformers import CLIPModel

        model = _lookup_model(model_id, CLIPModel.from_pretrained)
        return ts.ArrayType((model.config.projection_dim,), dtype=ts.FloatType(), nullable=False)
    except ImportError:
        return ts.ArrayType((None,), dtype=ts.FloatType(), nullable=False)


@pxt.udf(batch_size=4)
def detr_for_object_detection(
    image: Batch[PIL.Image.Image], *, model_id: str, threshold: float = 0.5, revision: str = 'no_timm'
) -> Batch[dict]:
    """
    Computes DETR object detections for the specified image. `model_id` should be a reference to a pretrained
    [DETR Model](https://huggingface.co/docs/transformers/model_doc/detr).

    __Requirements:__

    - `pip install torch transformers`

    Args:
        image: The image to embed.
        model_id: The pretrained model to use for object detection.

    Returns:
        A dictionary containing the output of the object detection model, in the following format:

            ```python
            {
                'scores': [0.99, 0.999],  # list of confidence scores for each detected object
                'labels': [25, 25],  # list of COCO class labels for each detected object
                'label_text': ['giraffe', 'giraffe'],  # corresponding text names of class labels
                'boxes': [[51.942, 356.174, 181.481, 413.975], [383.225, 58.66, 605.64, 361.346]]
                    # list of bounding boxes for each detected object, as [x1, y1, x2, y2]
            }
            ```

    Examples:
        Add a computed column that applies the model `facebook/detr-resnet-50` to an existing
        Pixeltable column `image` of the table `tbl`:

        >>> tbl.add_computed_column(detections=detr_for_object_detection(
        ...     tbl.image,
        ...     model_id='facebook/detr-resnet-50',
        ...     threshold=0.8
        ... ))
    """
    env.Env.get().require_package('transformers')
    device = resolve_torch_device('auto')
    import torch
    from transformers import DetrForObjectDetection, DetrImageProcessor

    model = _lookup_model(
        model_id, lambda x: DetrForObjectDetection.from_pretrained(x, revision=revision), device=device
    )
    processor = _lookup_processor(model_id, lambda x: DetrImageProcessor.from_pretrained(x, revision=revision))
    normalized_images = [normalize_image_mode(img) for img in image]

    with torch.no_grad():
        inputs = processor(images=normalized_images, return_tensors='pt')
        outputs = model(**inputs.to(device))
        results = processor.post_process_object_detection(
            outputs, threshold=threshold, target_sizes=[(img.height, img.width) for img in image]
        )

    return [
        {
            'scores': [score.item() for score in result['scores']],
            'labels': [label.item() for label in result['labels']],
            'label_text': [model.config.id2label[label.item()] for label in result['labels']],
            'boxes': [box.tolist() for box in result['boxes']],
        }
        for result in results
    ]


@pxt.udf(batch_size=4)
def vit_for_image_classification(
    image: Batch[PIL.Image.Image], *, model_id: str, top_k: int = 5
) -> Batch[dict[str, Any]]:
    """
    Computes image classifications for the specified image using a Vision Transformer (ViT) model.
    `model_id` should be a reference to a pretrained [ViT Model](https://huggingface.co/docs/transformers/en/model_doc/vit).

    __Note:__ Be sure the model is a ViT model that is trained for image classification (that is, a model designed for
    use with the
    [ViTForImageClassification](https://huggingface.co/docs/transformers/en/model_doc/vit#transformers.ViTForImageClassification)
    class), such as `google/vit-base-patch16-224`. General feature-extraction models such as
    `google/vit-base-patch16-224-in21k` will not produce the desired results.

    __Requirements:__

    - `pip install torch transformers`

    Args:
        image: The image to classify.
        model_id: The pretrained model to use for the classification.
        top_k: The number of classes to return.

    Returns:
        A dictionary containing the output of the image classification model, in the following format:

        ```python
        {
            'scores': [0.325, 0.198, 0.105],  # list of probabilities of the top-k most likely classes
            'labels': [340, 353, 386],  # list of class IDs for the top-k most likely classes
            'label_text': ['zebra', 'gazelle', 'African elephant, Loxodonta africana'],
                # corresponding text names of the top-k most likely classes
        ```

    Examples:
        Add a computed column that applies the model `google/vit-base-patch16-224` to an existing
        Pixeltable column `image` of the table `tbl`, returning the 10 most likely classes for each image:

        >>> tbl.add_computed_column(image_class=vit_for_image_classification(
        ...     tbl.image,
        ...     model_id='google/vit-base-patch16-224',
        ...     top_k=10
        ... ))
    """
    env.Env.get().require_package('transformers')
    device = resolve_torch_device('auto')
    import torch
    from transformers import ViTForImageClassification, ViTImageProcessor

    model: ViTForImageClassification = _lookup_model(model_id, ViTForImageClassification.from_pretrained, device=device)
    processor = _lookup_processor(model_id, ViTImageProcessor.from_pretrained)
    normalized_images = [normalize_image_mode(img) for img in image]

    with torch.no_grad():
        inputs = processor(images=normalized_images, return_tensors='pt')
        outputs = model(**inputs.to(device))
        logits = outputs.logits

    probs = torch.softmax(logits, dim=-1)
    top_k_probs, top_k_indices = torch.topk(probs, top_k, dim=-1)

    # There is no official post_process method for ViT models; for consistency, we structure the output
    # the same way as the output of the DETR model given by `post_process_object_detection`.
    return [
        {
            'scores': [top_k_probs[n, k].item() for k in range(top_k_probs.shape[1])],
            'labels': [top_k_indices[n, k].item() for k in range(top_k_probs.shape[1])],
            'label_text': [model.config.id2label[top_k_indices[n, k].item()] for k in range(top_k_probs.shape[1])],
        }
        for n in range(top_k_probs.shape[0])
    ]


@pxt.udf
def speech2text_for_conditional_generation(audio: pxt.Audio, *, model_id: str, language: Optional[str] = None) -> str:
    """
    Transcribes or translates speech to text using a Speech2Text model. `model_id` should be a reference to a
    pretrained [Speech2Text](https://huggingface.co/docs/transformers/en/model_doc/speech_to_text) model.

    __Requirements:__

    - `pip install torch torchaudio sentencepiece transformers`

    Args:
        audio: The audio clip to transcribe or translate.
        model_id: The pretrained model to use for the transcription or translation.
        language: If using a multilingual translation model, the language code to translate to. If not provided,
            the model's default language will be used. If the model is not translation model, is not a
            multilingual model, or does not support the specified language, an error will be raised.

    Returns:
        The transcribed or translated text.

    Examples:
        Add a computed column that applies the model `facebook/s2t-small-librispeech-asr` to an existing
        Pixeltable column `audio` of the table `tbl`:

        >>> tbl.add_computed_column(transcription=speech2text_for_conditional_generation(
        ...     tbl.audio,
        ...     model_id='facebook/s2t-small-librispeech-asr'
        ... ))

        Add a computed column that applies the model `facebook/s2t-medium-mustc-multilingual-st` to an existing
        Pixeltable column `audio` of the table `tbl`, translating the audio to French:

        >>> tbl.add_computed_column(translation=speech2text_for_conditional_generation(
        ...     tbl.audio,
        ...     model_id='facebook/s2t-medium-mustc-multilingual-st',
        ...     language='fr'
        ... ))
    """
    env.Env.get().require_package('transformers')
    env.Env.get().require_package('torchaudio')
    env.Env.get().require_package('sentencepiece')
    device = resolve_torch_device('auto', allow_mps=False)  # Doesn't seem to work on 'mps'; use 'cpu' instead
    import torch
    import torchaudio  # type: ignore[import-untyped]
    from transformers import Speech2TextForConditionalGeneration, Speech2TextProcessor, Speech2TextTokenizer

    model = _lookup_model(model_id, Speech2TextForConditionalGeneration.from_pretrained, device=device)
    processor = _lookup_processor(model_id, Speech2TextProcessor.from_pretrained)
    tokenizer = processor.tokenizer
    assert isinstance(processor, Speech2TextProcessor)
    assert isinstance(tokenizer, Speech2TextTokenizer)

    if language is not None and language not in tokenizer.lang_code_to_id:
        raise excs.Error(
            f"Language code '{language}' is not supported by the model '{model_id}'. "
            f'Supported languages are: {list(tokenizer.lang_code_to_id.keys())}'
        )

    forced_bos_token_id: Optional[int] = None if language is None else tokenizer.lang_code_to_id[language]

    # Get the model's sampling rate. Default to 16 kHz (the standard) if not in config
    model_sampling_rate = getattr(model.config, 'sampling_rate', 16_000)

    waveform, sampling_rate = torchaudio.load(audio)

    # Resample to the model's sampling rate, if necessary
    if sampling_rate != model_sampling_rate:
        waveform = torchaudio.transforms.Resample(sampling_rate, model_sampling_rate)(waveform)

    # Average the channels to get a single-channel waveform as a 1D tensor (if the original waveform is already
    # mono, this will simply squeeze the tensor)
    assert waveform.dim() == 2
    waveform = torch.mean(waveform, dim=0)
    assert waveform.dim() == 1

    with torch.no_grad():
        inputs = processor(waveform, sampling_rate=model_sampling_rate, return_tensors='pt')
        generated_ids = model.generate(**inputs.to(device), forced_bos_token_id=forced_bos_token_id).to('cpu')

    transcription = processor.batch_decode(generated_ids, skip_special_tokens=True)
    return transcription


@pxt.udf
def detr_to_coco(image: PIL.Image.Image, detr_info: dict[str, Any]) -> dict[str, Any]:
    """
    Converts the output of a DETR object detection model to COCO format.

    Args:
        image: The image for which detections were computed.
        detr_info: The output of a DETR object detection model, as returned by `detr_for_object_detection`.

    Returns:
        A dictionary containing the data from `detr_info`, converted to COCO format.

    Examples:
        Add a computed column that converts the output `tbl.detections` to COCO format, where `tbl.image`
        is the image for which detections were computed:

        >>> tbl.add_computed_column(detections_coco=detr_to_coco(tbl.image, tbl.detections))
    """
    bboxes, labels = detr_info['boxes'], detr_info['labels']
    annotations = [
        {'bbox': [bbox[0], bbox[1], bbox[2] - bbox[0], bbox[3] - bbox[1]], 'category': label}
        for bbox, label in zip(bboxes, labels)
    ]
    return {'image': {'width': image.width, 'height': image.height}, 'annotations': annotations}


T = TypeVar('T')


def _lookup_model(
    model_id: str, create: Callable[..., T], device: Optional[str] = None, pass_device_to_create: bool = False
) -> T:
    from torch import nn

    key = (model_id, create, device)  # For safety, include the `create` callable in the cache key
    if key not in _model_cache:
        if pass_device_to_create:
            model = create(model_id, device=device)
        else:
            model = create(model_id)
        if isinstance(model, nn.Module):
            if not pass_device_to_create and device is not None:
                model.to(device)
            model.eval()
        _model_cache[key] = model
    return _model_cache[key]


def _lookup_processor(model_id: str, create: Callable[[str], T]) -> T:
    key = (model_id, create)  # For safety, include the `create` callable in the cache key
    if key not in _processor_cache:
        _processor_cache[key] = create(model_id)
    return _processor_cache[key]


_model_cache: dict[tuple[str, Callable, Optional[str]], Any] = {}
_processor_cache: dict[tuple[str, Callable], Any] = {}


__all__ = local_public_names(__name__)


def __dir__() -> list[str]:
    return __all__



import logging
import sys
from typing import Any, Callable, Optional, TypeVar

R = TypeVar('R')


def _is_in_exception() -> bool:
    """
    Check if code is currently executing within an exception context.
    """
    current_exception = sys.exc_info()[1]
    return current_exception is not None


def run_cleanup_on_exception(cleanup_func: Callable[..., R], *args: Any, **kwargs: Any) -> Optional[R]:
    """
    Runs cleanup only when running in exception context.

    The function `run_cleanup_on_exception()` should be used to clean up resources when an operation fails.
    This is typically done using a try, except, and finally block, with the resource cleanup logic placed within
    the except block. However, this pattern may not handle KeyboardInterrupt exceptions.
    To ensure that resources are always cleaned up at least once when an exception or KeyboardInterrupt occurs,
    create an idempotent function for cleaning up resources and pass it to the `run_cleanup_on_exception()` function
    from the finally block.
    """
    if _is_in_exception():
        return run_cleanup(cleanup_func, *args, raise_error=False, **kwargs)
    return None


def run_cleanup(cleanup_func: Callable[..., R], *args: Any, raise_error: bool = True, **kwargs: Any) -> Optional[R]:
    """
    Runs a cleanup function. If interrupted, retry cleanup.
    The `run_cleanup()` function ensures that the `cleanup_func()` function executes at least once.
    If the `cleanup_func()` is interrupted during execution, it will be retried.

    Args:
        cleanup_func: an idempotent function
        raise_error: raise an exception if an error occurs during cleanup.
    """
    try:
        logging.debug(f'Running cleanup function: {cleanup_func.__name__!r}')
        return cleanup_func(*args, **kwargs)
    except KeyboardInterrupt as interrupt:
        # Save original exception and re-attempt cleanup
        original_exception = interrupt
        logging.debug(f'Cleanup {cleanup_func.__name__!r} interrupted, retrying')
        try:
            return cleanup_func(*args, **kwargs)
        except Exception as e:
            # Suppress this exception
            logging.error(f'Cleanup {cleanup_func.__name__!r} failed with exception {e}')
        raise KeyboardInterrupt from original_exception
    except Exception as e:
        logging.error(f'Cleanup {cleanup_func.__name__!r} failed with exception {e}')
        if raise_error:
            raise e
    return None



from __future__ import annotations

from typing import Any, NoReturn

from pixeltable import type_system as ts

from .data_row import DataRow
from .expr import Expr
from .row_builder import RowBuilder
from .sql_element_cache import SqlElementCache


class Variable(Expr):
    """An expr parameter, needed for ExprTemplateFunctions

    A Variable has a name and type and needs to have been replaced by an actual expression before evaluation.
    """

    def __init__(self, name: str, col_type: ts.ColumnType):
        super().__init__(col_type)
        self.name = name
        self.id = self._create_id()

    def _id_attrs(self) -> list[tuple[str, Any]]:
        return [*super()._id_attrs(), ('name', self.name)]

    def default_column_name(self) -> NoReturn:
        raise NotImplementedError()

    def _equals(self, other: Variable) -> bool:
        return self.name == other.name

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        return f"Variable('{self.name}')"

    def sql_expr(self, _: SqlElementCache) -> NoReturn:
        raise NotImplementedError()

    def eval(self, data_row: DataRow, row_builder: RowBuilder) -> NoReturn:
        raise NotImplementedError()

    def _as_dict(self) -> dict:
        return {'name': self.name, 'type': self.col_type.as_dict(), **super()._as_dict()}

    @classmethod
    def _from_dict(cls, d: dict, _: list[Expr]) -> Variable:
        return cls(d['name'], ts.ColumnType.from_dict(d['type']))



import logging
from uuid import UUID

import sqlalchemy as sql

from pixeltable.metadata import register_converter
from pixeltable.metadata.converters.util import convert_table_md

_logger = logging.getLogger('pixeltable')


@register_converter(version=23)
def _(engine: sql.engine.Engine) -> None:
    convert_table_md(engine, table_md_updater=__update_table_md)


def __update_table_md(table_md: dict, table_id: UUID) -> None:
    """update the index metadata to add indexed_col_tbl_id column if it is missing

    Args:
        table_md (dict): copy of the original table metadata. this gets updated in place.
        table_id (UUID): the table id

    """
    if len(table_md['index_md']) == 0:
        return
    for idx_md in table_md['index_md'].values():
        if 'indexed_col_tbl_id' not in idx_md:
            # index metadata is missing indexed_col_tbl_id
            # assume that the indexed column is in the same table
            # and update the index metadata.
            _logger.info(f'Updating index metadata for table: {table_id} index: {idx_md["id"]}')
            idx_md['indexed_col_tbl_id'] = str(table_id)



import json
from pathlib import Path
from typing import Any

import PIL

import pixeltable as pxt
import pixeltable.exceptions as excs

format_msg = """

Required format:
{
    'image': PIL.Image.Image,
    'annotations': [
        {
            'bbox': [x: int, y: int, w: int, h: int],
            'category': str | int,
        },
        ...
    ],
}
"""


def _verify_input_dict(input_dict: dict[str, Any]) -> None:
    """Verify that input_dict is a valid input dict for write_coco_dataset()"""
    if not isinstance(input_dict, dict):
        raise excs.Error(f'Expected dict, got {input_dict}{format_msg}')
    if 'image' not in input_dict:
        raise excs.Error(f'Missing key "image" in input dict: {input_dict}{format_msg}')
    if not isinstance(input_dict['image'], PIL.Image.Image):
        raise excs.Error(f'Value for "image" is not a PIL.Image.Image: {input_dict}{format_msg}')
    if 'annotations' not in input_dict:
        raise excs.Error(f'Missing key "annotations" in input dict: {input_dict}{format_msg}')
    if not isinstance(input_dict['annotations'], list):
        raise excs.Error(f'Value for "annotations" is not a list: {input_dict}{format_msg}')
    for annotation in input_dict['annotations']:
        if not isinstance(annotation, dict):
            raise excs.Error(f'Annotation is not a dict: {annotation}{format_msg}')
        if 'bbox' not in annotation:
            raise excs.Error(f'Missing key "bbox" in annotation: {annotation}{format_msg}')
        if not isinstance(annotation['bbox'], list):
            raise excs.Error(f'Value for "bbox" is not a list [x, y, w, h]: {annotation}{format_msg}')
        if len(annotation['bbox']) != 4 or not all(isinstance(x, int) for x in annotation['bbox']):
            raise excs.Error(f'Key "bbox" is not a list [x, y, w, h] of ints: {annotation}{format_msg}')
        if 'category' not in annotation:
            raise excs.Error(f'Missing key "category" in annotation: {annotation}{format_msg}')
        if not isinstance(annotation['category'], (str, int)):
            raise excs.Error(f'Value for "category" is not a str or int: {annotation}{format_msg}')


def write_coco_dataset(df: pxt.DataFrame, dest_path: Path) -> Path:
    """Export a DataFrame result set as a COCO dataset in dest_path and return the path of the data.json file."""
    # TODO: validate schema
    if len(df._select_list_exprs) != 1 or not df._select_list_exprs[0].col_type.is_json_type():
        raise excs.Error(f'Expected exactly one json-typed column in select list: {df._select_list_exprs}')
    input_dict_slot_idx = -1  # df._select_list_exprs[0].slot_idx isn't valid until _exec()

    # create output dir
    assert not dest_path.exists()
    dest_path.mkdir(parents=False)
    images_dir = dest_path / 'images'
    images_dir.mkdir()

    images: list[dict[str, Any]] = []
    img_id = -1
    annotations: list[dict[str, Any]] = []
    ann_id = -1
    categories: set[Any] = set()
    for input_row in df._exec():
        if input_dict_slot_idx == -1:
            input_dict_expr = df._select_list_exprs[0]
            input_dict_slot_idx = input_dict_expr.slot_idx
            input_dict = input_row[input_dict_slot_idx]
            _verify_input_dict(input_dict)

            # we want to know the slot idx of the image used in the input dict, so that we can check whether we
            # already have a local path for it
            input_dict_dependencies = input_dict_expr.dependencies()
            img_slot_idx = next((e.slot_idx for e in input_dict_dependencies if e.col_type.is_image_type()), None)
            assert img_slot_idx is not None
        else:
            input_dict = input_row[input_dict_slot_idx]
            _verify_input_dict(input_dict)

        # create image record
        img_id += 1

        # get a local path for the image
        img = input_dict['image']
        if input_row.file_paths[img_slot_idx] is not None:
            # we already have a local path
            img_path = Path(input_row.file_paths[img_slot_idx])
            # TODO: if the path leads to our tmp dir, we need to move the file
        else:
            # we need to create a local path
            img_path = images_dir / f'{img_id}.jpg'
            img.save(img_path)

        images.append({'id': img_id, 'file_name': str(img_path), 'width': img.width, 'height': img.height})

        # create annotation records for this image
        for annotation in input_dict['annotations']:
            ann_id += 1
            _, _, w, h = annotation['bbox']
            category = annotation['category']
            categories.add(category)
            annotations.append(
                {
                    'id': ann_id,
                    'image_id': img_id,
                    # we use the category name here and fix it up at the end, when we have assigned category ids
                    'category_id': category,
                    'bbox': annotation['bbox'],
                    'area': w * h,
                    'iscrowd': 0,
                }
            )

    # replace category names with ids
    category_ids = {category: id for id, category in enumerate(sorted(categories))}
    for annotation in annotations:
        annotation['category_id'] = category_ids[annotation['category_id']]

    result = {
        'images': images,
        'annotations': annotations,
        'categories': [{'id': id, 'name': category} for category, id in category_ids.items()],
    }
    output_path = dest_path / 'data.json'
    with open(output_path, 'w', encoding='utf-8') as fp:
        json.dump(result, fp)
    return output_path


COCO_2017_CATEGORIES = {
    0: 'N/A',
    1: 'person',
    2: 'bicycle',
    3: 'car',
    4: 'motorcycle',
    5: 'airplane',
    6: 'bus',
    7: 'train',
    8: 'truck',
    9: 'boat',
    10: 'traffic light',
    11: 'fire hydrant',
    12: 'N/A',
    13: 'stop sign',
    14: 'parking meter',
    15: 'bench',
    16: 'bird',
    17: 'cat',
    18: 'dog',
    19: 'horse',
    20: 'sheep',
    21: 'cow',
    22: 'elephant',
    23: 'bear',
    24: 'zebra',
    25: 'giraffe',
    26: 'N/A',
    27: 'backpack',
    28: 'umbrella',
    29: 'N/A',
    30: 'N/A',
    31: 'handbag',
    32: 'tie',
    33: 'suitcase',
    34: 'frisbee',
    35: 'skis',
    36: 'snowboard',
    37: 'sports ball',
    38: 'kite',
    39: 'baseball bat',
    40: 'baseball glove',
    41: 'skateboard',
    42: 'surfboard',
    43: 'tennis racket',
    44: 'bottle',
    45: 'N/A',
    46: 'wine glass',
    47: 'cup',
    48: 'fork',
    49: 'knife',
    50: 'spoon',
    51: 'bowl',
    52: 'banana',
    53: 'apple',
    54: 'sandwich',
    55: 'orange',
    56: 'broccoli',
    57: 'carrot',
    58: 'hot dog',
    59: 'pizza',
    60: 'donut',
    61: 'cake',
    62: 'chair',
    63: 'couch',
    64: 'potted plant',
    65: 'bed',
    66: 'N/A',
    67: 'dining table',
    68: 'N/A',
    69: 'N/A',
    70: 'toilet',
    71: 'N/A',
    72: 'tv',
    73: 'laptop',
    74: 'mouse',
    75: 'remote',
    76: 'keyboard',
    77: 'cell phone',
    78: 'microwave',
    79: 'oven',
    80: 'toaster',
    81: 'sink',
    82: 'refrigerator',
    83: 'N/A',
    84: 'book',
    85: 'clock',
    86: 'vase',
    87: 'scissors',
    88: 'teddy bear',
    89: 'hair drier',
    90: 'toothbrush',
}



from __future__ import annotations

import dataclasses
import enum
from textwrap import dedent
from typing import Any, Iterable, Literal, Optional, Sequence
from uuid import UUID

import sqlalchemy as sql

import pixeltable as pxt
from pixeltable import catalog, exceptions as excs, exec, exprs
from pixeltable.catalog import Column, TableVersionHandle
from pixeltable.exec.sql_node import OrderByClause, OrderByItem, combine_order_by_clauses, print_order_by_clause


def _is_agg_fn_call(e: exprs.Expr) -> bool:
    return isinstance(e, exprs.FunctionCall) and e.is_agg_fn_call and not e.is_window_fn_call


def _get_combined_ordering(
    o1: list[tuple[exprs.Expr, bool]], o2: list[tuple[exprs.Expr, bool]]
) -> list[tuple[exprs.Expr, bool]]:
    """Returns an ordering that's compatible with both o1 and o2, or an empty list if no such ordering exists"""
    result: list[tuple[exprs.Expr, bool]] = []
    # determine combined ordering
    for (e1, asc1), (e2, asc2) in zip(o1, o2):
        if e1.id != e2.id:
            return []
        if asc1 is not None and asc2 is not None and asc1 != asc2:
            return []
        asc = asc1 if asc1 is not None else asc2
        result.append((e1, asc))

    # add remaining ordering of the longer list
    prefix_len = min(len(o1), len(o2))
    if len(o1) > prefix_len:
        result.extend(o1[prefix_len:])
    elif len(o2) > prefix_len:
        result.extend(o2[prefix_len:])
    return result


class JoinType(enum.Enum):
    INNER = 0
    LEFT = 1
    # TODO: implement
    # RIGHT = 2
    FULL_OUTER = 3
    CROSS = 4

    LiteralType = Literal['inner', 'left', 'full_outer', 'cross']

    @classmethod
    def validated(cls, name: str, error_prefix: str) -> JoinType:
        try:
            return cls[name.upper()]
        except KeyError as exc:
            val_strs = ', '.join(f'{s.lower()!r}' for s in cls.__members__)
            raise excs.Error(f'{error_prefix} must be one of: [{val_strs}]') from exc


@dataclasses.dataclass
class JoinClause:
    """Corresponds to a single 'JOIN ... ON (...)' clause in a SELECT statement; excludes the joined table."""

    join_type: JoinType
    join_predicate: Optional[exprs.Expr]  # None for join_type == CROSS


@dataclasses.dataclass
class FromClause:
    """Corresponds to the From-clause ('FROM <tbl> JOIN ... ON (...) JOIN ...') of a SELECT statement"""

    tbls: list[catalog.TableVersionPath]
    join_clauses: list[JoinClause] = dataclasses.field(default_factory=list)

    @property
    def _first_tbl(self) -> catalog.TableVersionPath:
        assert len(self.tbls) == 1
        return self.tbls[0]


@dataclasses.dataclass
class SampleClause:
    """Defines a sampling clause for a table."""

    version: Optional[int]
    n: Optional[int]
    n_per_stratum: Optional[int]
    fraction: Optional[float]
    seed: Optional[int]
    stratify_exprs: Optional[list[exprs.Expr]]

    # This seed value is used if one is not supplied
    DEFAULT_SEED = 0

    # The version of the hashing algorithm used for ordering and fractional sampling.
    CURRENT_VERSION = 1

    def __post_init__(self) -> None:
        """If no version was provided, provide the default version"""
        if self.version is None:
            self.version = self.CURRENT_VERSION
        if self.seed is None:
            self.seed = self.DEFAULT_SEED

    @property
    def is_stratified(self) -> bool:
        """Check if the sampling is stratified"""
        return self.stratify_exprs is not None and len(self.stratify_exprs) > 0

    @property
    def is_repeatable(self) -> bool:
        """Return true if the same rows will continue to be sampled if source rows are added or deleted."""
        return not self.is_stratified and self.fraction is not None

    def display_str(self, inline: bool = False) -> str:
        return str(self)

    def as_dict(self) -> dict:
        """Return a dictionary representation of the object"""
        d = dataclasses.asdict(self)
        d['_classname'] = self.__class__.__name__
        if self.is_stratified:
            d['stratify_exprs'] = [e.as_dict() for e in self.stratify_exprs]
        return d

    @classmethod
    def from_dict(cls, d: dict) -> SampleClause:
        """Create a SampleClause from a dictionary representation"""
        d_cleaned = {key: value for key, value in d.items() if key != '_classname'}
        s = cls(**d_cleaned)
        if s.is_stratified:
            s.stratify_exprs = [exprs.Expr.from_dict(e) for e in d_cleaned.get('stratify_exprs', [])]
        return s

    def __repr__(self) -> str:
        s = ','.join(e.display_str(inline=True) for e in self.stratify_exprs)
        return (
            f'sample_{self.version}(n={self.n}, n_per_stratum={self.n_per_stratum}, '
            f'fraction={self.fraction}, seed={self.seed}, [{s}])'
        )

    @classmethod
    def fraction_to_md5_hex(cls, fraction: float) -> str:
        """Return the string representation of an approximation (to ~1e-9) of a fraction of the total space
        of md5 hash values.
        This is used for fractional sampling.
        """
        # Maximum count for the upper 32 bits of MD5: 2^32
        max_md5_value = (2**32) - 1

        # Calculate the fraction of this value
        threshold_int = max_md5_value * int(1_000_000_000 * fraction) // 1_000_000_000

        # Convert to hexadecimal string with padding
        return format(threshold_int, '08x') + 'ffffffffffffffffffffffff'


class Analyzer:
    """
    Performs semantic analysis of a query and stores the analysis state.
    """

    from_clause: FromClause
    all_exprs: list[exprs.Expr]  # union of all exprs, aside from sql_where_clause
    select_list: list[exprs.Expr]
    group_by_clause: Optional[list[exprs.Expr]]  # None for non-aggregate queries; [] for agg query w/o grouping
    grouping_exprs: list[exprs.Expr]  # [] for non-aggregate queries or agg query w/o grouping
    order_by_clause: OrderByClause
    stratify_exprs: list[exprs.Expr]  # [] if no stratiifcation is required
    sample_clause: Optional[SampleClause]  # None if no sampling clause is present

    sql_elements: exprs.SqlElementCache

    # Where clause of the Select stmt of the SQL scan
    sql_where_clause: Optional[exprs.Expr]

    # filter predicate applied to output rows of the SQL scan
    filter: Optional[exprs.Expr]

    agg_fn_calls: list[exprs.FunctionCall]  # grouping aggregation (ie, not window functions)
    window_fn_calls: list[exprs.FunctionCall]
    agg_order_by: list[exprs.Expr]

    def __init__(
        self,
        from_clause: FromClause,
        select_list: Sequence[exprs.Expr],
        where_clause: Optional[exprs.Expr] = None,
        group_by_clause: Optional[list[exprs.Expr]] = None,
        order_by_clause: Optional[list[tuple[exprs.Expr, bool]]] = None,
        sample_clause: Optional[SampleClause] = None,
    ):
        if order_by_clause is None:
            order_by_clause = []
        self.from_clause = from_clause
        self.sql_elements = exprs.SqlElementCache()

        # remove references to unstored computed cols
        self.select_list = [e.resolve_computed_cols() for e in select_list]
        if where_clause is not None:
            where_clause = where_clause.resolve_computed_cols()
        self.group_by_clause = (
            [e.resolve_computed_cols() for e in group_by_clause] if group_by_clause is not None else None
        )
        self.sample_clause = sample_clause
        if self.sample_clause is not None and self.sample_clause.is_stratified:
            self.stratify_exprs = [e.resolve_computed_cols() for e in sample_clause.stratify_exprs]
        else:
            self.stratify_exprs = []
        self.order_by_clause = [OrderByItem(e.resolve_computed_cols(), asc) for e, asc in order_by_clause]

        self.sql_where_clause = None
        self.filter = None
        if where_clause is not None:
            where_clause_conjuncts, self.filter = where_clause.split_conjuncts(self.sql_elements.contains)
            self.sql_where_clause = exprs.CompoundPredicate.make_conjunction(where_clause_conjuncts)

        # all exprs that are evaluated in Python; not executable
        self.all_exprs = self.select_list.copy()
        for join_clause in from_clause.join_clauses:
            if join_clause.join_predicate is not None:
                self.all_exprs.append(join_clause.join_predicate)
        if self.group_by_clause is not None:
            self.all_exprs.extend(self.group_by_clause)
        self.all_exprs.extend(self.stratify_exprs)
        self.all_exprs.extend(e for e, _ in self.order_by_clause)
        if self.filter is not None:
            if sample_clause is not None:
                raise excs.Error(f'Filter {self.filter} not expressible in SQL')
            self.all_exprs.append(self.filter)

        self.agg_order_by = []
        self.agg_fn_calls = []
        self.window_fn_calls = []
        self._analyze_agg()
        self.grouping_exprs = self.group_by_clause if self.group_by_clause is not None else []

    def _analyze_agg(self) -> None:
        """Check semantic correctness of aggregation and fill in agg-specific fields of Analyzer"""
        candidates = self.select_list
        agg_fn_calls = exprs.ExprSet(
            exprs.Expr.list_subexprs(
                candidates,
                expr_class=exprs.FunctionCall,
                filter=lambda e: bool(e.is_agg_fn_call and not e.is_window_fn_call),
            )
        )
        self.agg_fn_calls = list(agg_fn_calls)
        window_fn_calls = exprs.ExprSet(
            exprs.Expr.list_subexprs(
                candidates, expr_class=exprs.FunctionCall, filter=lambda e: bool(e.is_window_fn_call)
            )
        )
        self.window_fn_calls = list(window_fn_calls)
        if len(self.agg_fn_calls) == 0:
            # nothing to do
            return
        # if we're doing grouping aggregation and don't have an explicit Group By clause, we're creating a single group
        if self.group_by_clause is None:
            self.group_by_clause = []

        # check that select list only contains aggregate output
        grouping_expr_ids = {e.id for e in self.group_by_clause}
        is_agg_output = [self._determine_agg_status(e, grouping_expr_ids)[0] for e in self.select_list]
        if is_agg_output.count(False) > 0:
            raise excs.Error(
                f'Invalid non-aggregate expression in aggregate query: {self.select_list[is_agg_output.index(False)]}'
            )

        # check that Where clause and filter doesn't contain aggregates
        if self.sql_where_clause is not None and any(
            _is_agg_fn_call(e) for e in self.sql_where_clause.subexprs(expr_class=exprs.FunctionCall)
        ):
            raise excs.Error(f'where() cannot contain aggregate functions: {self.sql_where_clause}')
        if self.filter is not None and any(
            _is_agg_fn_call(e) for e in self.filter.subexprs(expr_class=exprs.FunctionCall)
        ):
            raise excs.Error(f'where() cannot contain aggregate functions: {self.filter}')

        # check that grouping exprs don't contain aggregates and can be expressed as SQL (we perform sort-based
        # aggregation and rely on the SqlScanNode returning data in the correct order)
        for e in self.group_by_clause:
            if not self.sql_elements.contains(e):
                raise excs.Error(f'Invalid grouping expression, needs to be expressible in SQL: {e}')
            if e._contains(filter=_is_agg_fn_call):
                raise excs.Error(f'Grouping expression contains aggregate function: {e}')

    def _determine_agg_status(self, e: exprs.Expr, grouping_expr_ids: set[int]) -> tuple[bool, bool]:
        """Determine whether expr is the input to or output of an aggregate function.
        Returns:
            (<is output>, <is input>)
        """
        if e.id in grouping_expr_ids:
            return True, True
        elif _is_agg_fn_call(e):
            for c in e.components:
                _, is_input = self._determine_agg_status(c, grouping_expr_ids)
                if not is_input:
                    raise excs.Error(f'Invalid nested aggregates: {e}')
            return True, False
        elif isinstance(e, exprs.Literal):
            return True, True
        elif isinstance(e, (exprs.ColumnRef, exprs.RowidRef)):
            # we already know that this isn't a grouping expr
            return False, True
        else:
            # an expression such as <grouping expr 1> + <grouping expr 2> can both be the output and input of agg
            assert len(e.components) > 0
            component_is_output, component_is_input = zip(
                *[self._determine_agg_status(c, grouping_expr_ids) for c in e.components]
            )
            is_output = component_is_output.count(True) == len(e.components)
            is_input = component_is_input.count(True) == len(e.components)
            if not is_output and not is_input:
                raise excs.Error(f'Invalid expression, mixes aggregate with non-aggregate: {e}')
            return is_output, is_input

    def finalize(self, row_builder: exprs.RowBuilder) -> None:
        """Make all exprs executable
        TODO: add EvalCtx for each expr list?
        """
        # maintain original composition of select list
        row_builder.set_slot_idxs(self.select_list, remove_duplicates=False)
        if self.group_by_clause is not None:
            row_builder.set_slot_idxs(self.group_by_clause)
        order_by_exprs = [e for e, _ in self.order_by_clause]
        row_builder.set_slot_idxs(order_by_exprs)
        row_builder.set_slot_idxs(self.all_exprs)
        if self.filter is not None:
            row_builder.set_slot_idxs([self.filter])
        row_builder.set_slot_idxs(self.agg_fn_calls)
        row_builder.set_slot_idxs(self.agg_order_by)

    def get_window_fn_ob_clause(self) -> Optional[OrderByClause]:
        clause: list[OrderByClause] = []
        for fn_call in self.window_fn_calls:
            # window functions require ordering by the group_by/order_by clauses
            group_by_exprs, order_by_exprs = fn_call.get_window_sort_exprs()
            clause.append(
                [OrderByItem(e, None) for e in group_by_exprs] + [OrderByItem(e, True) for e in order_by_exprs]
            )
        return combine_order_by_clauses(clause)

    def has_agg(self) -> bool:
        """True if there is any kind of aggregation in the query"""
        return self.group_by_clause is not None or len(self.agg_fn_calls) > 0 or len(self.window_fn_calls) > 0


class Planner:
    # TODO: create an exec.CountNode and change this to create_count_plan()
    @classmethod
    def create_count_stmt(cls, tbl: catalog.TableVersionPath, where_clause: Optional[exprs.Expr] = None) -> sql.Select:
        stmt = sql.select(sql.func.count().label('all_count'))
        refd_tbl_ids: set[UUID] = set()
        if where_clause is not None:
            analyzer = cls.analyze(tbl, where_clause)
            if analyzer.filter is not None:
                raise excs.Error(f'Filter {analyzer.filter} not expressible in SQL')
            clause_element = analyzer.sql_where_clause.sql_expr(analyzer.sql_elements)
            assert clause_element is not None
            stmt = stmt.where(clause_element)
            refd_tbl_ids = where_clause.tbl_ids()
        stmt = exec.SqlScanNode.create_from_clause(tbl, stmt, refd_tbl_ids)
        return stmt

    @classmethod
    def create_insert_plan(
        cls, tbl: catalog.TableVersion, rows: list[dict[str, Any]], ignore_errors: bool
    ) -> exec.ExecNode:
        """Creates a plan for TableVersion.insert()"""
        assert not tbl.is_view
        # stored_cols: all cols we need to store, incl computed cols (and indices)
        stored_cols = [c for c in tbl.cols_by_id.values() if c.is_stored]
        assert len(stored_cols) > 0  # there needs to be something to store

        cls.__check_valid_columns(tbl, stored_cols, 'inserted into')

        row_builder = exprs.RowBuilder([], stored_cols, [], tbl)

        # create InMemoryDataNode for 'rows'
        plan: exec.ExecNode = exec.InMemoryDataNode(
            TableVersionHandle(tbl.id, tbl.effective_version), rows, row_builder, tbl.next_row_id
        )

        media_input_col_info = [
            exprs.ColumnSlotIdx(col_ref.col, col_ref.slot_idx)
            for col_ref in row_builder.input_exprs
            if isinstance(col_ref, exprs.ColumnRef) and col_ref.col_type.is_media_type()
        ]
        if len(media_input_col_info) > 0:
            # prefetch external files for all input column refs
            plan = exec.CachePrefetchNode(tbl.id, media_input_col_info, input=plan)

        computed_exprs = row_builder.output_exprs - row_builder.input_exprs
        if len(computed_exprs) > 0:
            # add an ExprEvalNode when there are exprs to compute
            plan = exec.ExprEvalNode(
                row_builder, computed_exprs, plan.output_exprs, input=plan, maintain_input_order=False
            )

        stored_col_info = row_builder.output_slot_idxs()
        stored_img_col_info = [info for info in stored_col_info if info.col.col_type.is_image_type()]
        plan.set_stored_img_cols(stored_img_col_info)
        plan.set_ctx(
            exec.ExecContext(
                row_builder,
                batch_size=0,
                show_pbar=True,
                num_computed_exprs=len(computed_exprs),
                ignore_errors=ignore_errors,
            )
        )
        return plan

    @classmethod
    def rowid_columns(cls, target: TableVersionHandle, num_rowid_cols: Optional[int] = None) -> list[exprs.Expr]:
        """Return list of RowidRef for the given number of associated rowids"""
        if num_rowid_cols is None:
            num_rowid_cols = target.get().num_rowid_columns()
        return [exprs.RowidRef(target, i) for i in range(num_rowid_cols)]

    @classmethod
    def create_df_insert_plan(
        cls, tbl: catalog.TableVersion, df: 'pxt.DataFrame', ignore_errors: bool
    ) -> exec.ExecNode:
        assert not tbl.is_view
        plan = df._create_query_plan()  # ExecNode constructed by the DataFrame

        # Modify the plan RowBuilder to register the output columns
        for col_name, expr in zip(df.schema.keys(), df._select_list_exprs):
            assert col_name in tbl.cols_by_name
            col = tbl.cols_by_name[col_name]
            plan.row_builder.add_table_column(col, expr.slot_idx)

        stored_col_info = plan.row_builder.output_slot_idxs()
        stored_img_col_info = [info for info in stored_col_info if info.col.col_type.is_image_type()]
        plan.set_stored_img_cols(stored_img_col_info)

        plan.set_ctx(
            exec.ExecContext(
                plan.row_builder, batch_size=0, show_pbar=True, num_computed_exprs=0, ignore_errors=ignore_errors
            )
        )
        plan.ctx.num_rows = 0  # Unknown

        return plan

    @classmethod
    def create_update_plan(
        cls,
        tbl: catalog.TableVersionPath,
        update_targets: dict[catalog.Column, exprs.Expr],
        recompute_targets: list[catalog.Column],
        where_clause: Optional[exprs.Expr],
        cascade: bool,
    ) -> tuple[exec.ExecNode, list[str], list[catalog.Column]]:
        """Creates a plan to materialize updated rows.
        The plan:
        - retrieves rows that are visible at the current version of the table
        - materializes all stored columns and the update targets
        - if cascade is True, recomputes all computed columns that transitively depend on the updated columns
          and copies the values of all other stored columns
        - if cascade is False, copies all columns that aren't update targets from the original rows
        Returns:
            - root node of the plan
            - list of qualified column names that are getting updated
            - list of user-visible columns that are being recomputed
        """
        # retrieve all stored cols and all target exprs
        assert isinstance(tbl, catalog.TableVersionPath)
        target = tbl.tbl_version.get()  # the one we need to update
        updated_cols = list(update_targets.keys())
        recomputed_cols: set[Column]
        if len(recompute_targets) > 0:
            assert len(update_targets) == 0
            recomputed_cols = {*recompute_targets}
            if cascade:
                recomputed_cols |= target.get_dependent_columns(recomputed_cols)
        else:
            recomputed_cols = target.get_dependent_columns(updated_cols) if cascade else set()
        # regardless of cascade, we need to update all indices on any updated/recomputed column
        idx_val_cols = target.get_idx_val_columns(set(updated_cols) | recomputed_cols)
        recomputed_cols.update(idx_val_cols)
        # we only need to recompute stored columns (unstored ones are substituted away)
        recomputed_cols = {c for c in recomputed_cols if c.is_stored}

        cls.__check_valid_columns(tbl.tbl_version.get(), recomputed_cols, 'updated in')

        recomputed_base_cols = {col for col in recomputed_cols if col.tbl.id == tbl.tbl_version.id}
        copied_cols = [
            col
            for col in target.cols_by_id.values()
            if col.is_stored and col not in updated_cols and col not in recomputed_base_cols
        ]
        select_list: list[exprs.Expr] = [exprs.ColumnRef(col) for col in copied_cols]
        select_list.extend(update_targets.values())

        recomputed_exprs = [
            c.value_expr.copy().resolve_computed_cols(resolve_cols=recomputed_base_cols) for c in recomputed_base_cols
        ]
        # recomputed cols reference the new values of the updated cols
        spec: dict[exprs.Expr, exprs.Expr] = {exprs.ColumnRef(col): e for col, e in update_targets.items()}
        exprs.Expr.list_substitute(recomputed_exprs, spec)
        select_list.extend(recomputed_exprs)

        # we need to retrieve the PK columns of the existing rows
        plan = cls.create_query_plan(FromClause(tbls=[tbl]), select_list, where_clause=where_clause, ignore_errors=True)
        all_base_cols = copied_cols + updated_cols + list(recomputed_base_cols)  # same order as select_list
        # update row builder with column information
        for i, col in enumerate(all_base_cols):
            plan.row_builder.add_table_column(col, select_list[i].slot_idx)
        plan.ctx.num_computed_exprs = len(recomputed_exprs)
        recomputed_user_cols = [c for c in recomputed_cols if c.name is not None]
        return plan, [f'{c.tbl.name}.{c.name}' for c in updated_cols + recomputed_user_cols], recomputed_user_cols

    @classmethod
    def __check_valid_columns(
        cls, tbl: catalog.TableVersion, cols: Iterable[Column], op_name: Literal['inserted into', 'updated in']
    ) -> None:
        for col in cols:
            if col.value_expr is not None and not col.value_expr.is_valid:
                raise excs.Error(
                    dedent(
                        f"""
                        Data cannot be {op_name} the table {tbl.name!r},
                        because the column {col.name!r} is currently invalid:
                        {{validation_error}}
                        """
                    )
                    .strip()
                    .format(validation_error=col.value_expr.validation_error)
                )

    @classmethod
    def create_batch_update_plan(
        cls,
        tbl: catalog.TableVersionPath,
        batch: list[dict[catalog.Column, exprs.Expr]],
        rowids: list[tuple[int, ...]],
        cascade: bool,
    ) -> tuple[exec.ExecNode, exec.RowUpdateNode, sql.ColumnElement[bool], list[catalog.Column], list[catalog.Column]]:
        """
        Returns:
        - root node of the plan to produce the updated rows
        - RowUpdateNode of plan
        - Where clause for deleting the current versions of updated rows
        - list of columns that are getting updated
        - list of user-visible columns that are being recomputed
        """
        assert isinstance(tbl, catalog.TableVersionPath)
        target = tbl.tbl_version.get()  # the one we need to update
        sa_key_cols: list[sql.Column] = []
        key_vals: list[tuple] = []
        if len(rowids) > 0:
            sa_key_cols = target.store_tbl.rowid_columns()
            key_vals = rowids
        else:
            pk_cols = target.primary_key_columns()
            sa_key_cols = [c.sa_col for c in pk_cols]
            key_vals = [tuple(row[col].val for col in pk_cols) for row in batch]

        # retrieve all stored cols and all target exprs
        updated_cols = batch[0].keys() - target.primary_key_columns()
        recomputed_cols = target.get_dependent_columns(updated_cols) if cascade else set()
        # regardless of cascade, we need to update all indices on any updated column
        idx_val_cols = target.get_idx_val_columns(updated_cols)
        recomputed_cols.update(idx_val_cols)
        # we only need to recompute stored columns (unstored ones are substituted away)
        recomputed_cols = {c for c in recomputed_cols if c.is_stored}
        recomputed_base_cols = {col for col in recomputed_cols if col.tbl.id == target.id}
        copied_cols = [
            col
            for col in target.cols_by_id.values()
            if col.is_stored and col not in updated_cols and col not in recomputed_base_cols
        ]
        select_list: list[exprs.Expr] = [exprs.ColumnRef(col) for col in copied_cols]
        select_list.extend(exprs.ColumnRef(col) for col in updated_cols)

        recomputed_exprs = [
            c.value_expr.copy().resolve_computed_cols(resolve_cols=recomputed_base_cols) for c in recomputed_base_cols
        ]
        # the RowUpdateNode updates columns in-place, ie, in the original ColumnRef; no further substitution is needed
        select_list.extend(recomputed_exprs)

        # ExecNode tree (from bottom to top):
        # - SqlLookupNode to retrieve the existing rows
        # - RowUpdateNode to update the retrieved rows
        # - ExprEvalNode to evaluate the remaining output exprs
        analyzer = Analyzer(FromClause(tbls=[tbl]), select_list)
        sql_exprs = list(
            exprs.Expr.list_subexprs(analyzer.all_exprs, filter=analyzer.sql_elements.contains, traverse_matches=False)
        )
        row_builder = exprs.RowBuilder(analyzer.all_exprs, [], sql_exprs, target)
        analyzer.finalize(row_builder)
        sql_lookup_node = exec.SqlLookupNode(tbl, row_builder, sql_exprs, sa_key_cols, key_vals)
        col_vals = [{col: row[col].val for col in updated_cols} for row in batch]
        row_update_node = exec.RowUpdateNode(tbl, key_vals, len(rowids) > 0, col_vals, row_builder, sql_lookup_node)
        plan: exec.ExecNode = row_update_node
        if not cls._is_contained_in(analyzer.select_list, sql_exprs):
            # we need an ExprEvalNode to evaluate the remaining output exprs
            plan = exec.ExprEvalNode(row_builder, analyzer.select_list, sql_exprs, input=plan)
        # update row builder with column information
        all_base_cols = copied_cols + list(updated_cols) + list(recomputed_base_cols)  # same order as select_list
        row_builder.set_slot_idxs(select_list, remove_duplicates=False)
        for i, col in enumerate(all_base_cols):
            plan.row_builder.add_table_column(col, select_list[i].slot_idx)
        ctx = exec.ExecContext(row_builder, num_computed_exprs=len(recomputed_exprs))
        # we're returning everything to the user, so we might as well do it in a single batch
        ctx.batch_size = 0
        plan.set_ctx(ctx)
        recomputed_user_cols = [c for c in recomputed_cols if c.name is not None]
        return (
            plan,
            row_update_node,
            sql_lookup_node.where_clause_element,
            list(updated_cols) + recomputed_user_cols,
            recomputed_user_cols,
        )

    @classmethod
    def create_view_update_plan(
        cls, view: catalog.TableVersionPath, recompute_targets: list[catalog.Column]
    ) -> exec.ExecNode:
        """Creates a plan to materialize updated rows for a view, given that the base table has been updated.
        The plan:
        - retrieves rows that are visible at the current version of the table and satisfy the view predicate
        - materializes all stored columns and the update targets
        - if cascade is True, recomputes all computed columns that transitively depend on the updated columns
          and copies the values of all other stored columns
        - if cascade is False, copies all columns that aren't update targets from the original rows

        TODO: unify with create_view_load_plan()

        Returns:
            - root node of the plan
            - list of qualified column names that are getting updated
            - list of columns that are being recomputed
        """
        assert isinstance(view, catalog.TableVersionPath)
        assert view.is_view
        target = view.tbl_version.get()  # the one we need to update
        # retrieve all stored cols and all target exprs
        recomputed_cols = set(recompute_targets.copy())
        copied_cols = [col for col in target.cols_by_id.values() if col.is_stored and col not in recomputed_cols]
        select_list: list[exprs.Expr] = [exprs.ColumnRef(col) for col in copied_cols]
        # resolve recomputed exprs to stored columns in the base
        recomputed_exprs = [
            c.value_expr.copy().resolve_computed_cols(resolve_cols=recomputed_cols) for c in recomputed_cols
        ]
        select_list.extend(recomputed_exprs)

        # we need to retrieve the PK columns of the existing rows
        plan = cls.create_query_plan(
            FromClause(tbls=[view]),
            select_list,
            where_clause=target.predicate,
            ignore_errors=True,
            exact_version_only=view.get_bases(),
        )
        plan.ctx.num_computed_exprs = len(recomputed_exprs)
        for i, col in enumerate(copied_cols + list(recomputed_cols)):  # same order as select_list
            plan.row_builder.add_table_column(col, select_list[i].slot_idx)
        # TODO: avoid duplication with view_load_plan() logic (where does this belong?)
        stored_img_col_info = [
            info for info in plan.row_builder.output_slot_idxs() if info.col.col_type.is_image_type()
        ]
        plan.set_stored_img_cols(stored_img_col_info)
        return plan

    @classmethod
    def create_view_load_plan(
        cls, view: catalog.TableVersionPath, propagates_insert: bool = False
    ) -> tuple[exec.ExecNode, int]:
        """Creates a query plan for populating a view.

        Args:
            view: the view to populate
            propagates_insert: if True, we're propagating a base update to this view

        Returns:
            - root node of the plan
            - number of materialized values per row
        """
        assert isinstance(view, catalog.TableVersionPath)
        assert view.is_view
        # things we need to materialize as DataRows:
        # 1. stored computed cols
        # - iterator columns are effectively computed, just not with a value_expr
        # - we can ignore stored non-computed columns because they have a default value that is supplied directly by
        #   the store
        target = view.tbl_version.get()  # the one we need to populate
        stored_cols = [c for c in target.cols_by_id.values() if c.is_stored]
        # 2. for component views: iterator args
        iterator_args = [target.iterator_args] if target.iterator_args is not None else []

        from_clause = FromClause(tbls=[view.base])
        base_analyzer = Analyzer(
            from_clause, iterator_args, where_clause=target.predicate, sample_clause=target.sample_clause
        )
        row_builder = exprs.RowBuilder(base_analyzer.all_exprs, stored_cols, [], target)

        # if we're propagating an insert, we only want to see those base rows that were created for the current version
        # execution plan:
        # 1. materialize exprs computed from the base that are needed for stored view columns
        # 2. if it's an iterator view, expand the base rows into component rows
        # 3. materialize stored view columns that haven't been produced by step 1
        base_output_exprs = [e for e in row_builder.default_eval_ctx.exprs if e.is_bound_by([view.base])]
        view_output_exprs = [
            e
            for e in row_builder.default_eval_ctx.target_exprs
            if e.is_bound_by([view]) and not e.is_bound_by([view.base])
        ]

        # Create a new analyzer reflecting exactly what is required from the base table
        base_analyzer = Analyzer(
            from_clause, base_output_exprs, where_clause=target.predicate, sample_clause=target.sample_clause
        )
        base_eval_ctx = row_builder.create_eval_ctx(base_analyzer.all_exprs)
        plan = cls._create_query_plan(
            row_builder=row_builder,
            analyzer=base_analyzer,
            eval_ctx=base_eval_ctx,
            with_pk=True,
            exact_version_only=view.get_bases() if propagates_insert else [],
        )
        exec_ctx = plan.ctx
        if target.is_component_view:
            plan = exec.ComponentIterationNode(view.tbl_version, plan)
        if len(view_output_exprs) > 0:
            plan = exec.ExprEvalNode(
                row_builder, output_exprs=view_output_exprs, input_exprs=base_output_exprs, input=plan
            )

        stored_img_col_info = [info for info in row_builder.output_slot_idxs() if info.col.col_type.is_image_type()]
        plan.set_stored_img_cols(stored_img_col_info)
        exec_ctx.ignore_errors = True
        plan.set_ctx(exec_ctx)
        return plan, len(row_builder.default_eval_ctx.target_exprs)

    @classmethod
    def _verify_join_clauses(cls, analyzer: Analyzer) -> None:
        """Verify that join clauses are expressible in SQL"""
        for join_clause in analyzer.from_clause.join_clauses:
            if join_clause.join_predicate is not None and analyzer.sql_elements.get(join_clause.join_predicate) is None:
                raise excs.Error(f'Join predicate {join_clause.join_predicate} not expressible in SQL')

    @classmethod
    def _create_combined_ordering(cls, analyzer: Analyzer, verify_agg: bool) -> Optional[OrderByClause]:
        """Verify that the various ordering requirements don't conflict and return a combined ordering"""
        ob_clauses: list[OrderByClause] = [analyzer.order_by_clause.copy()]

        if verify_agg:
            ordering: OrderByClause
            for fn_call in analyzer.window_fn_calls:
                # window functions require ordering by the group_by/order_by clauses
                gb, ob = fn_call.get_window_sort_exprs()
                ordering = [OrderByItem(e, None) for e in gb] + [OrderByItem(e, True) for e in ob]
                ob_clauses.append(ordering)
            for fn_call in analyzer.agg_fn_calls:
                # agg functions with an ordering requirement are implicitly ascending
                ordering = [OrderByItem(e, None) for e in analyzer.group_by_clause] + [
                    OrderByItem(e, True) for e in fn_call.get_agg_order_by()
                ]
                ob_clauses.append(ordering)

        if len(ob_clauses) == 0:
            return None
        elif len(ob_clauses) == 1:
            return ob_clauses[0]

        combined_ordering = ob_clauses[0]
        for ordering in ob_clauses[1:]:
            combined = combine_order_by_clauses([combined_ordering, ordering])
            if combined is None:
                raise excs.Error(
                    f'Incompatible ordering requirements: '
                    f'{print_order_by_clause(combined_ordering)} vs {print_order_by_clause(ordering)}'
                )
            combined_ordering = combined
        return combined_ordering

    @classmethod
    def _is_contained_in(cls, l1: Iterable[exprs.Expr], l2: Iterable[exprs.Expr]) -> bool:
        """Returns True if l1 is contained in l2"""
        return {e.id for e in l1} <= {e.id for e in l2}

    @classmethod
    def _insert_prefetch_node(
        cls, tbl_id: UUID, row_builder: exprs.RowBuilder, input_node: exec.ExecNode
    ) -> exec.ExecNode:
        """Returns a CachePrefetchNode into the plan if needed, otherwise returns input"""
        # we prefetch external files for all media ColumnRefs, even those that aren't part of the dependencies
        # of output_exprs: if unstored iterator columns are present, we might need to materialize ColumnRefs that
        # aren't explicitly captured as dependencies
        media_col_refs = [
            e for e in list(row_builder.unique_exprs) if isinstance(e, exprs.ColumnRef) and e.col_type.is_media_type()
        ]
        if len(media_col_refs) == 0:
            return input_node
        # we need to prefetch external files for media column types
        file_col_info = [exprs.ColumnSlotIdx(e.col, e.slot_idx) for e in media_col_refs]
        prefetch_node = exec.CachePrefetchNode(tbl_id, file_col_info, input_node)
        return prefetch_node

    @classmethod
    def create_query_plan(
        cls,
        from_clause: FromClause,
        select_list: Optional[list[exprs.Expr]] = None,
        where_clause: Optional[exprs.Expr] = None,
        group_by_clause: Optional[list[exprs.Expr]] = None,
        order_by_clause: Optional[list[tuple[exprs.Expr, bool]]] = None,
        limit: Optional[exprs.Expr] = None,
        sample_clause: Optional[SampleClause] = None,
        ignore_errors: bool = False,
        exact_version_only: Optional[list[catalog.TableVersionHandle]] = None,
    ) -> exec.ExecNode:
        """Return plan for executing a query.
        Updates 'select_list' in place to make it executable.
        TODO: make exact_version_only a flag and use the versions from tbl
        """
        if select_list is None:
            select_list = []
        if order_by_clause is None:
            order_by_clause = []
        if exact_version_only is None:
            exact_version_only = []

        analyzer = Analyzer(
            from_clause,
            select_list,
            where_clause=where_clause,
            group_by_clause=group_by_clause,
            order_by_clause=order_by_clause,
            sample_clause=sample_clause,
        )
        # If the from_clause has a single table, we can use it as the context table for the RowBuilder.
        # Otherwise there is no context table, but that's ok, because the context table is only needed for
        # table mutations, which can't happen during a join.
        context_tbl = from_clause.tbls[0].tbl_version.get() if len(from_clause.tbls) == 1 else None
        row_builder = exprs.RowBuilder(analyzer.all_exprs, [], [], context_tbl)

        analyzer.finalize(row_builder)
        # select_list: we need to materialize everything that's been collected
        # with_pk: for now, we always retrieve the PK, because we need it for the file cache
        eval_ctx = row_builder.create_eval_ctx(analyzer.select_list)
        plan = cls._create_query_plan(
            row_builder=row_builder,
            analyzer=analyzer,
            eval_ctx=eval_ctx,
            limit=limit,
            with_pk=True,
            exact_version_only=exact_version_only,
        )
        plan.ctx.ignore_errors = ignore_errors
        select_list.clear()
        select_list.extend(analyzer.select_list)
        return plan

    @classmethod
    def _create_query_plan(
        cls,
        row_builder: exprs.RowBuilder,
        analyzer: Analyzer,
        eval_ctx: exprs.RowBuilder.EvalCtx,
        limit: Optional[exprs.Expr] = None,
        with_pk: bool = False,
        exact_version_only: Optional[list[catalog.TableVersionHandle]] = None,
    ) -> exec.ExecNode:
        """
        Create plan to materialize eval_ctx.

        Args:
            plan_target: if not None, generate a plan that materializes only expression that can be evaluted
                in the context of that table version (eg, if 'tbl' is a view, 'plan_target' might be the base)
        TODO: make exact_version_only a flag and use the versions from tbl
        """
        if exact_version_only is None:
            exact_version_only = []
        sql_elements = analyzer.sql_elements
        is_python_agg = not sql_elements.contains_all(analyzer.agg_fn_calls) or not sql_elements.contains_all(
            analyzer.window_fn_calls
        )
        ctx = exec.ExecContext(row_builder)
        combined_ordering = cls._create_combined_ordering(analyzer, verify_agg=is_python_agg)
        cls._verify_join_clauses(analyzer)

        # materialized with SQL table scans (ie, single-table SELECT statements):
        # - select list subexprs that aren't aggregates
        # - join clause subexprs
        # - subexprs of Where clause conjuncts that can't be run in SQL
        # - all grouping exprs
        # - all stratify exprs
        candidates = list(
            exprs.Expr.list_subexprs(
                analyzer.select_list,
                filter=lambda e: (
                    sql_elements.contains(e)
                    and not e._contains(cls=exprs.FunctionCall, filter=lambda e: bool(e.is_agg_fn_call))
                ),
                traverse_matches=False,
            )
        )
        if analyzer.filter is not None:
            candidates.extend(
                exprs.Expr.subexprs(analyzer.filter, filter=sql_elements.contains, traverse_matches=False)
            )
        candidates.extend(
            exprs.Expr.list_subexprs(analyzer.grouping_exprs, filter=sql_elements.contains, traverse_matches=False)
        )
        candidates.extend(
            exprs.Expr.list_subexprs(analyzer.stratify_exprs, filter=sql_elements.contains, traverse_matches=False)
        )
        # not isinstance(...): we don't want to materialize Literals via a Select
        sql_exprs = exprs.ExprSet(e for e in candidates if not isinstance(e, exprs.Literal))

        # create table scans; each scan produces subexprs of (sql_exprs + join clauses)
        join_exprs = exprs.ExprSet(
            join_clause.join_predicate
            for join_clause in analyzer.from_clause.join_clauses
            if join_clause.join_predicate is not None
        )
        scan_target_exprs = sql_exprs | join_exprs
        tbl_scan_plans: list[exec.SqlScanNode] = []
        plan: exec.ExecNode
        for tbl in analyzer.from_clause.tbls:
            # materialize all subexprs of scan_target_exprs that are bound by tbl
            tbl_scan_exprs = exprs.ExprSet(
                exprs.Expr.list_subexprs(
                    scan_target_exprs,
                    filter=lambda e: e.is_bound_by([tbl]) and not isinstance(e, exprs.Literal),
                    traverse_matches=False,
                )
            )
            plan = exec.SqlScanNode(
                tbl, row_builder, select_list=tbl_scan_exprs, set_pk=with_pk, exact_version_only=exact_version_only
            )
            tbl_scan_plans.append(plan)

        if len(analyzer.from_clause.join_clauses) > 0:
            plan = exec.SqlJoinNode(
                row_builder,
                inputs=tbl_scan_plans,
                join_clauses=analyzer.from_clause.join_clauses,
                select_list=sql_exprs,
            )
        else:
            plan = tbl_scan_plans[0]

        if analyzer.sql_where_clause is not None:
            plan.set_where(analyzer.sql_where_clause)
        if analyzer.filter is not None:
            plan.set_py_filter(analyzer.filter)
        if len(analyzer.window_fn_calls) > 0:
            # we need to order the input for window functions
            plan.set_order_by(analyzer.get_window_fn_ob_clause())

        if analyzer.sample_clause is not None:
            plan = exec.SqlSampleNode(
                row_builder,
                input=plan,
                select_list=tbl_scan_exprs,
                sample_clause=analyzer.sample_clause,
                stratify_exprs=analyzer.stratify_exprs,
            )

        plan = cls._insert_prefetch_node(tbl.tbl_version.id, row_builder, plan)

        if analyzer.group_by_clause is not None:
            # we're doing grouping aggregation; the input of the AggregateNode are the grouping exprs plus the
            # args of the agg fn calls
            agg_input = exprs.ExprSet(analyzer.grouping_exprs.copy())
            for fn_call in analyzer.agg_fn_calls:
                agg_input.update(fn_call.components)
            if not sql_exprs.issuperset(agg_input):
                # we need an ExprEvalNode
                plan = exec.ExprEvalNode(row_builder, agg_input, sql_exprs, input=plan)

            # batch size for aggregation input: this could be the entire table, so we need to divide it into
            # smaller batches; at the same time, we need to make the batches large enough to amortize the
            # function call overhead
            ctx.batch_size = 16

            # do aggregation in SQL if all agg exprs can be translated
            if (
                sql_elements.contains_all(analyzer.select_list)
                and sql_elements.contains_all(analyzer.grouping_exprs)
                and isinstance(plan, exec.SqlNode)
                and plan.to_cte() is not None
            ):
                plan = exec.SqlAggregationNode(
                    row_builder, input=plan, select_list=analyzer.select_list, group_by_items=analyzer.group_by_clause
                )
            else:
                input_sql_node = plan.get_node(exec.SqlNode)
                assert combined_ordering is not None
                input_sql_node.set_order_by(combined_ordering)
                plan = exec.AggregationNode(
                    tbl.tbl_version,
                    row_builder,
                    analyzer.group_by_clause,
                    analyzer.agg_fn_calls + analyzer.window_fn_calls,
                    agg_input,
                    input=plan,
                )
                typecheck_dummy = analyzer.grouping_exprs + analyzer.agg_fn_calls + analyzer.window_fn_calls
                agg_output = exprs.ExprSet(typecheck_dummy)
                if not agg_output.issuperset(exprs.ExprSet(eval_ctx.target_exprs)):
                    # we need an ExprEvalNode to evaluate the remaining output exprs
                    plan = exec.ExprEvalNode(row_builder, eval_ctx.target_exprs, agg_output, input=plan)
        else:
            if not exprs.ExprSet(sql_exprs).issuperset(exprs.ExprSet(eval_ctx.target_exprs)):
                # we need an ExprEvalNode to evaluate the remaining output exprs
                plan = exec.ExprEvalNode(row_builder, eval_ctx.target_exprs, sql_exprs, input=plan)
            # we're returning everything to the user, so we might as well do it in a single batch
            # TODO: return smaller batches in order to increase inter-ExecNode parallelism
            ctx.batch_size = 0

        sql_node = plan.get_node(exec.SqlNode)
        if len(analyzer.order_by_clause) > 0:
            # we have the last SqlNode we created produce the ordering
            assert sql_node is not None
            sql_node.set_order_by(analyzer.order_by_clause)

        # if we don't need an ordered result, tell the ExprEvalNode not to maintain input order (which allows us to
        # return batches earlier)
        if sql_node is not None and len(sql_node.order_by_clause) == 0:
            expr_eval_node = plan.get_node(exec.ExprEvalNode)
            if expr_eval_node is not None:
                expr_eval_node.set_input_order(False)

        if limit is not None:
            assert isinstance(limit, exprs.Literal)
            plan.set_limit(limit.val)

        plan.set_ctx(ctx)
        return plan

    @classmethod
    def analyze(cls, tbl: catalog.TableVersionPath, where_clause: exprs.Expr) -> Analyzer:
        return Analyzer(FromClause(tbls=[tbl]), [], where_clause=where_clause)

    @classmethod
    def create_add_column_plan(cls, tbl: catalog.TableVersionPath, col: catalog.Column) -> exec.ExecNode:
        """Creates a plan for InsertableTable.add_column()
        Returns:
            plan: the plan to execute
            value_expr slot idx for the plan output (for computed cols)
        """
        assert isinstance(tbl, catalog.TableVersionPath)
        row_builder = exprs.RowBuilder(output_exprs=[], columns=[col], input_exprs=[], tbl=tbl.tbl_version.get())
        analyzer = Analyzer(FromClause(tbls=[tbl]), row_builder.default_eval_ctx.target_exprs)
        plan = cls._create_query_plan(
            row_builder=row_builder, analyzer=analyzer, eval_ctx=row_builder.default_eval_ctx, with_pk=True
        )
        plan.ctx.batch_size = 16
        plan.ctx.show_pbar = True
        plan.ctx.ignore_errors = True
        computed_exprs = row_builder.output_exprs - row_builder.input_exprs
        plan.ctx.num_computed_exprs = len(computed_exprs)  # we are adding a computed column, so we need to evaluate it

        # we want to flush images
        if col.is_computed and col.is_stored and col.col_type.is_image_type():
            plan.set_stored_img_cols(row_builder.output_slot_idxs())
        return plan



from __future__ import annotations

from typing import Any, Optional, Union

import jmespath
import sqlalchemy as sql

from pixeltable import catalog, exceptions as excs, type_system as ts

from .data_row import DataRow
from .expr import Expr
from .globals import print_slice
from .json_mapper import JsonMapperDispatch
from .object_ref import ObjectRef
from .row_builder import RowBuilder
from .sql_element_cache import SqlElementCache


class JsonPath(Expr):
    def __init__(
        self, anchor: Optional[Expr], path_elements: Optional[list[Union[str, int, slice]]] = None, scope_idx: int = 0
    ) -> None:
        """
        anchor can be None, in which case this is a relative JsonPath and the anchor is set later via set_anchor().
        scope_idx: for relative paths, index of referenced JsonMapper
        (0: indicates the immediately preceding JsonMapper, -1: the parent of the immediately preceding mapper, ...)
        """
        if path_elements is None:
            path_elements = []
        super().__init__(ts.JsonType(nullable=True))  # JsonPath expressions are always nullable
        if anchor is not None:
            self.components = [anchor]
        self.path_elements: list[Union[str, int, slice]] = path_elements
        self.compiled_path = jmespath.compile(self._json_path()) if len(path_elements) > 0 else None
        self.scope_idx = scope_idx
        # NOTE: the _create_id() result will change if set_anchor() gets called;
        # this is not a problem, because _create_id() shouldn't be called after init()
        self.id = self._create_id()

    def __repr__(self) -> str:
        # else 'R': the anchor is RELATIVE_PATH_ROOT
        anchor_str = str(self._anchor) if self._anchor is not None else 'R'
        if len(self.path_elements) == 0:
            return anchor_str
        return f'{anchor_str}{"." if isinstance(self.path_elements[0], str) else ""}{self._json_path()}'

    def _as_dict(self) -> dict:
        assert len(self.components) <= 1
        components_dict: dict[str, Any]
        if len(self.components) == 0 or isinstance(self.components[0], ObjectRef):
            # If the anchor is an ObjectRef, it means this JsonPath is a bound relative path. We store it as a relative
            # path, *not* a bound path (which has no meaning in the dict).
            components_dict = {}
        else:
            components_dict = super()._as_dict()
        path_elements = [[el.start, el.stop, el.step] if isinstance(el, slice) else el for el in self.path_elements]
        return {'path_elements': path_elements, 'scope_idx': self.scope_idx, **components_dict}

    @classmethod
    def _from_dict(cls, d: dict, components: list[Expr]) -> JsonPath:
        assert 'path_elements' in d
        assert 'scope_idx' in d
        assert len(components) <= 1
        anchor = components[0] if len(components) == 1 else None
        path_elements = [slice(el[0], el[1], el[2]) if isinstance(el, list) else el for el in d['path_elements']]
        return cls(anchor, path_elements, d['scope_idx'])

    @property
    def _anchor(self) -> Optional[Expr]:
        return None if len(self.components) == 0 else self.components[0]

    def set_anchor(self, anchor: Expr) -> None:
        assert len(self.components) == 0
        self.components = [anchor]

    def is_relative_path(self) -> bool:
        return self._anchor is None

    def _has_relative_path(self) -> bool:
        return self.is_relative_path() or super()._has_relative_path()

    def _bind_rel_paths(self, mapper: Optional['JsonMapperDispatch'] = None) -> None:
        if self.is_relative_path():
            # TODO: take scope_idx into account
            self.set_anchor(mapper.scope_anchor)
        else:
            self._anchor._bind_rel_paths(mapper)

    def __call__(self, *args: object, **kwargs: object) -> 'JsonPath':
        """
        Construct a relative path that references an ancestor of the immediately enclosing JsonMapper.
        """
        if not self.is_relative_path():
            raise excs.Error('() for an absolute path is invalid')
        if len(args) != 1 or not isinstance(args[0], int) or args[0] >= 0:
            raise excs.Error('R() requires a negative index')
        return JsonPath(None, [], args[0])

    def __getattr__(self, name: str) -> 'JsonPath':
        assert isinstance(name, str)
        return JsonPath(self._anchor, [*self.path_elements, name])

    def __getitem__(self, index: object) -> 'JsonPath':
        if isinstance(index, (int, slice, str)):
            return JsonPath(self._anchor, [*self.path_elements, index])
        raise excs.Error(f'Invalid json list index: {index}')

    def default_column_name(self) -> Optional[str]:
        anchor_name = self._anchor.default_column_name() if self._anchor is not None else ''
        ret_name = f'{anchor_name}.{self._json_path()}'

        def cleanup_char(s: str) -> str:
            if s == '.':
                return '_'
            elif s == '*':
                return 'star'
            elif s.isalnum():
                return s
            else:
                return ''

        clean_name = ''.join(map(cleanup_char, ret_name))
        clean_name = clean_name.lstrip('_')  # remove leading underscore
        if not clean_name:  # Replace '' with None
            clean_name = None

        assert clean_name is None or catalog.is_valid_identifier(clean_name)
        return clean_name

    def _equals(self, other: JsonPath) -> bool:
        return self.path_elements == other.path_elements

    def _id_attrs(self) -> list[tuple[str, Any]]:
        return [*super()._id_attrs(), ('path_elements', self.path_elements)]

    def sql_expr(self, _: SqlElementCache) -> Optional[sql.ColumnElement]:
        """
        Postgres appears to have a bug: jsonb_path_query('{a: [{b: 0}, {b: 1}]}', '$.a.b') returns
        *two* rows (each containing col val 0), not a single row with [0, 0].
        We need to use a workaround: retrieve the entire dict, then use jmespath to extract the path correctly.
        """
        # path_str = '$.' + '.'.join(self.path_elements)
        # assert isinstance(self._anchor(), ColumnRef)
        # return sql.func.jsonb_path_query(self._anchor().col.sa_col, path_str)
        return None

    def _json_path(self) -> str:
        assert len(self.path_elements) > 0
        result: list[str] = []
        for element in self.path_elements:
            if element == '*':
                result.append('[*]')
            elif isinstance(element, str):
                result.append(f'{"." if len(result) > 0 else ""}{element}')
            elif isinstance(element, int):
                result.append(f'[{element}]')
            elif isinstance(element, slice):
                result.append(f'[{print_slice(element)}]')
        return ''.join(result)

    def eval(self, data_row: DataRow, row_builder: RowBuilder) -> None:
        assert self._anchor is not None, self
        val = data_row[self._anchor.slot_idx]
        if self.compiled_path is not None:
            val = self.compiled_path.search(val)
        data_row[self.slot_idx] = val


RELATIVE_PATH_ROOT = JsonPath(None)



import sqlalchemy as sql

from pixeltable.metadata import register_converter
from pixeltable.metadata.schema import Table, TableSchemaVersion


@register_converter(version=10)
def _(engine: sql.engine.Engine) -> None:
    default_table_attrs = {'comment': None, 'num_retained_versions': 10}
    with engine.begin() as conn:
        # Because `parameters` wasn't actually used for anything,
        # we can simply delete it without any data loss.
        conn.execute(sql.update(Table).values(md=Table.md - 'parameters'))
        # Add `table_attrs` to all instances of tableschemaversions.md.
        conn.execute(sql.update(TableSchemaVersion).values(md=TableSchemaVersion.md.concat(default_table_attrs)))
